# original developer: Henning Glawe, henning.glawe@mpsd.mpg.de
# updated by: temok.salazar@physik.hu-berlin.de

import logging
import re
import copy


LOGGER = logging.getLogger(__name__)


def parse_qe_xc_num(xc_functional_num):
    """parse Quantum Espresso XC number/index notation
    :returns: list with 6 integer elements (components of QE XC functionals)
              [iexch, icorr, igcx, igcc, imeta, inlc]
                iexch, icorr - density exchange/correlation
                igcx, igcc   - gradient correction exchange/correlation
                imeta        - meta-GGA
                inlc         - non-local term of Van der Waals functionals
    """
    xf_num_split_i = []
    xf_num_split = re.split(r'\s+', xc_functional_num.strip())
    if len(xf_num_split) > 6:
        # tested versions of espresso <=5.4.0 have 6 elements
        raise RuntimeError("unsupported number of XC components: %d" % (
            len(xf_num_split)))
    elif len(xf_num_split) == 6:
        # trivial case: we got 6 components from simply splitting
        xf_num_split_i = [int(x) for x in xf_num_split]
    elif len(xf_num_split) == 1 and len(xc_functional_num) == 4:
        # old 4-component form without space separator
        xf_num_split_i = [int(x) for x in re.findall(
            '(\d)', xc_functional_num)]
    elif len(xc_functional_num) == 10:
        # intermediate versions used 2-digit, 5 component form,
        # occasionally missing spaces
        # examples:
        # ( 1 4 4 0 1)
        # ( 0 413 0 2)  <- missing space
        xf_num_split_i = [int(x) for x in re.findall(
            '([ \d]\d)', xc_functional_num)]
    else:
        try:
            xf_num_split_i = [int(x) for x in xc_functional_num.split(' ') if x.strip() != '']
        except Exception:
            raise RuntimeError("unparsable input: '%s'", xc_functional_num)

    if len(xf_num_split_i) < 1:
        raise RuntimeError("this should not happen")
    # zero-pad up to 6 elements
    xf_num_split_i += [0] * (6 - len(xf_num_split_i))
    return xf_num_split_i


def translate_qe_xc_num(xc_functional_num, exact_exchange_fraction=None):
    # all remainders of excact_exchange_fraction must be ordinary DFT
    if exact_exchange_fraction is None:
        exact_exchange_fraction = 0.0
    xf_num = parse_qe_xc_num(xc_functional_num)
    LOGGER.debug('num <- input: %s <- %s, exx_fraction: %s', str(xf_num),
                 xc_functional_num, str(exact_exchange_fraction))
    # use dictionaries to ensure uniqueness:
    #   exchange/correlation functionals may be combined into _XC_, and we
    #   only want to emit such combinations once
    xc_data = {}
    #   libXC definitions of GGAs/Hybrids include the density contribution
    #   while QE allows to freely combine density, gradient etc. dependency
    #   so we need to remove at least the default settings to be compliant
    #   with libXC and NOMAD metaInfo
    xc_data_remove = {}
    # collect everything that will go to section_method
    xc_section_method = {}
    for component_i in range(6):
        this_xf_num = xf_num[component_i]
        if this_xf_num == 0:
            # 0 means unset component
            continue
        component_max = len(XC_COMPONENT[component_i])-1
        this_component = None
        if this_xf_num > component_max:
            LOGGER.error(
                "%s[%d] beyond limit of %d",
                XC_COMPONENT_NAME[component_i], this_xf_num, component_max)
        else:
            this_component = XC_COMPONENT[component_i][this_xf_num]
        if this_component is None:
            raise RuntimeError("Undefined XC component %s[%d]" % (
                XC_COMPONENT_NAME[component_i], this_xf_num))
        xc_section_method.update(this_component['xc_section_method'])
        for this_term in this_component['xc_terms']:
            apply_term_add(
                xc_data, this_term, exact_exchange_fraction)
        # collect all to-be-removed terms
        for this_term in this_component.get('xc_terms_remove', []):
            apply_term_add(
                xc_data_remove, this_term, exact_exchange_fraction)
    # remove all collected terms from xc_terms_remove from xc_data
    apply_terms_remove(xc_data, xc_data_remove)
    # filter terms in xc_data:
    apply_terms_filter(xc_data)
    xc_functional = xc_functional_str(xc_data)
    # apply shortcuts/overrides for libXC compliance
    if xc_functional in LIBXC_SHORTCUT:
        xc_data = {}
        for this_term in LIBXC_SHORTCUT[xc_functional]['xc_terms']:
            apply_term_add(
                xc_data, this_term, exact_exchange_fraction)
        apply_terms_filter(xc_data)
        xc_functional = xc_functional_str(xc_data)
    xc_section_method['XC_functional'] = xc_functional
    result = []
    for k in sorted(xc_data.keys()):
        v = xc_data[k]
        result.append(v)
    return (xc_section_method, result)


def apply_term_add(xc_data, this_term, exact_exchange_fraction):
    term = copy.deepcopy(this_term)
    if 'exx_compute_weight' in term:
        term['XC_functional_weight'] = term['exx_compute_weight'](
            exact_exchange_fraction)
    if 'XC_functional_weight' not in term:
        term['XC_functional_weight'] = 1.0
    if term['XC_functional_name'] not in xc_data:
        xc_data[term['XC_functional_name']] = term
    else:
        LOGGER.info("pre-existing XC term: %s",
                    term['XC_functional_name'])
    return xc_data


def apply_terms_remove(xc_data, xc_data_remove):
    for (k, v) in xc_data_remove.items():
        if k in xc_data:
            xc_data[k]['XC_functional_weight'] -= v['XC_functional_weight']
        else:
            xc_data[k] = v
            xc_data[k]['XC_functional_weight'] *= -1.0


def apply_terms_filter(xc_data):
    for (k, v) in list(xc_data.items()):
        if abs(v['XC_functional_weight']) < 0.01:
            del xc_data[k]
        else:
            if abs(v['XC_functional_weight'] - 1.0) < 0.01:
                del v['XC_functional_weight']
            v.pop('exx_compute_weight', None)


def xc_functional_str(xc_data, separator='+'):
    result = ''
    for k in sorted(xc_data.keys()):
        v = xc_data[k]
        if len(result) > 0 and v.get('XC_functional_weight', 1.0) > 0:
            result += separator
        if v.get('XC_functional_weight', None) is not None:
            result += '%.3f*' % (v['XC_functional_weight'])
        result += v['XC_functional_name']
    return result











# origin: espresso-5.4.0/Modules/funct.f90
# update:
# . New exchange-correlation functionals exist in
# .     espresso-6.5.0/Modules/funct.f90
#   short comments mark the corresponding new metainfo
EXCHANGE = [
    None,
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name':       'sla',
            'x_qe_xc_iexch_comment':    'Slater (alpha=2/3)',
            'x_qe_xc_iexch':      1,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_X',
            'XC_functional_parameters': {'alpha': 1.0},
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name':       'sl1',
            'x_qe_xc_iexch_comment':    'Slater (alpha=1.0)',
            'x_qe_xc_iexch':      2,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'x_qe_LDA_X_RELATIVISTIC',
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name':       'rxc',
            'x_qe_xc_iexch_comment':    'Relativistic Slater',
            'x_qe_xc_iexch':      3,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'OEP_EXX',
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name':       'oep',
            'x_qe_xc_iexch_comment':    'Optimized Effective Potential',
            'x_qe_xc_iexch':      4,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'HF_X',
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name':       'hf',
            'x_qe_xc_iexch_comment':    'Hartree-Fock',
            'x_qe_xc_iexch':      5,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'HF_X',
            'exx_compute_weight': lambda exx: exx,
            'XC_functional_weight': 0.25,
        }, {
            'XC_functional_name': 'LDA_X',
            'exx_compute_weight': lambda exx: (1.0 - exx),
            'XC_functional_weight': 0.75,
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name':       "pb0x",
            'x_qe_xc_iexch_comment':    'PBE0 (Slater*0.75+HF*0.25)',
            'x_qe_xc_iexch':      6,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'HF_X',
            'exx_compute_weight': lambda exx: exx,
            'XC_functional_weight': 0.20,
        }, {
            'XC_functional_name': 'LDA_X',
            'exx_compute_weight': lambda exx: (1.0 - exx),
            'XC_functional_weight': 0.8,
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name':       "b3lp",
            'x_qe_xc_iexch_comment':    "B3LYP(Slater*0.80+HF*0.20)",
            'x_qe_xc_iexch':      7,
        },
    },
    # LDA_X_KZK is not part of libXC. Look up it at
    # 'https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/XC-functional'
    {
        'xc_terms': [{
            'XC_functional_name': "LDA_X_KZK",
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name':       "kzk",
            'x_qe_xc_iexch_comment':    "Finite-size corrections",
            'x_qe_xc_iexch':      8,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'HF_X',
            'exx_compute_weight': lambda exx: exx,
            'XC_functional_weight': 0.218,
        }, {
            'XC_functional_name': 'LDA_X',
            'exx_compute_weight': lambda exx: (1.0 - exx),
            'XC_functional_weight': 0.782,
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name':       "x3lp",
            'x_qe_xc_iexch_comment':    "X3LYP(Slater*0.782+HF*0.218)",
            'x_qe_xc_iexch':      9,
        },
    },
    # update for espresso-6.5.0: KLI
        {
        'xc_terms': [{
            'XC_functional_name': "LDA_X_KLI",
        }],
        'xc_section_method': {
            'x_qe_xc_iexch_name': "kli",
            'x_qe_xc_iexch_comment': "KLI aproximation for exx",
            'x_qe_xc_iexch': 10,
        },
    },
]

# Correlation functionals UNchanged between espresso v5.4 & v6.5
CORRELATION = [
    None,
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_PZ',
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "pz",
            'x_qe_xc_icorr_comment':    "Perdew-Zunger",
            'x_qe_xc_icorr':      1,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_VWN',
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "vwn",
            'x_qe_xc_icorr_comment':    "Vosko-Wilk-Nusair",
            'x_qe_xc_icorr':      2,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_LYP',
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "lyp",
            'x_qe_xc_icorr_comment':    "Lee-Yang-Parr",
            'x_qe_xc_icorr':      3,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "LDA_C_PW",
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "pw",
            'x_qe_xc_icorr_comment':    "Perdew-Wang",
            'x_qe_xc_icorr':      4,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_WIGNER',
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "wig",
            'x_qe_xc_icorr_comment':    "Wigner",
            'x_qe_xc_icorr':      5,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_HL',
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "hl",
            'x_qe_xc_icorr_comment':    "Hedin-Lunqvist",
            'x_qe_xc_icorr':      6,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_OB_PZ',
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "obz",
            'x_qe_xc_icorr_comment':    "Ortiz-Ballone form for PZ",
            'x_qe_xc_icorr':      7,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_OB_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "obw",
            'x_qe_xc_icorr_comment':    "Ortiz-Ballone form for PW",
            'x_qe_xc_icorr':      8,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_GL',
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "gl",
            'x_qe_xc_icorr_comment':    "Gunnarson-Lunqvist",
            'x_qe_xc_icorr':      9,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "LDA_C_KZK",
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "kzk",
            'x_qe_xc_icorr_comment':    "Finite-size corrections",
            'x_qe_xc_icorr':      10,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_VWN_RPA',
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "vwn-rpa",
            'x_qe_xc_icorr_comment':    "Vosko-Wilk-Nusair, alt param",
            'x_qe_xc_icorr':      11,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_VWN',
            'XC_functional_weight': 0.19,
        }, {
            'XC_functional_name': 'LDA_C_LYP',
            'XC_functional_weight': 0.81,
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "b3lp",
            'x_qe_xc_icorr_comment':    "B3LYP (0.19*vwn+0.81*lyp)",
            'x_qe_xc_icorr':      12,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_VWN_RPA',
            'XC_functional_weight': 0.19,
        }, {
            'XC_functional_name': 'LDA_C_LYP',
            'XC_functional_weight': 0.81,
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':    "b3lpv1r",
            'x_qe_xc_icorr_comment': "B3LYP-VWN-1-RPA (0.19*vwn_rpa+0.81*lyp)",
            'x_qe_xc_icorr':         13,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': 'LDA_C_VWN_RPA',
            'XC_functional_weight': 0.129,
        }, {
            'XC_functional_name': 'LDA_C_LYP',
            'XC_functional_weight': 0.871,
        }],
        'xc_section_method': {
            'x_qe_xc_icorr_name':       "x3lp",
            'x_qe_xc_icorr_comment':    "X3LYP (0.129*vwn_rpa+0.871*lyp)",
            'x_qe_xc_icorr':      14,
        },
    },
]

# New "exchange_gradient_correction" functionals for q-espresso (qe) v6.5
#    igcx=[1..28] unchanged between qe-v5.4 & v6.5
# New additions:  igcx=[29..42]

EXCHANGE_GRADIENT_CORRECTION = [
    None,
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_B88",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "b88",
            'x_qe_xc_igcx_comment':    "Becke88 (beta=0.0042)",
            'x_qe_xc_igcx':      1,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_PW91",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "ggx",
            'x_qe_xc_igcx_comment':    "Perdew-Wang 91",
            'x_qe_xc_igcx':      2,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_PBE",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "pbx",
            'x_qe_xc_igcx_comment':    "Perdew-Burke-Ernzenhof exch",
            'x_qe_xc_igcx':      3,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_PBE_R",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "rpb",
            'x_qe_xc_igcx_comment':    "revised PBE by Zhang-Yang",
            'x_qe_xc_igcx':      4,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_XC_HCTH_120",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "hcth",
            'x_qe_xc_igcx_comment':    "Cambridge exch, Handy et al, HCTH/120",
            'x_qe_xc_igcx':      5,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_OPTX",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "optx",
            'x_qe_xc_igcx_comment':    "Handy's exchange functional",
            'x_qe_xc_igcx':      6,
        },
    },
    {
        # igcx=7 is not defined in 5.4's funct.f90
        #        definition taken from 5.0, which did not have separate imeta
        'xc_terms': [{
            'XC_functional_name': "MGGA_X_TPSS",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':    "tpss",
            'x_qe_xc_igcx_comment': "TPSS Meta-GGA (Espresso-version < 5.1)",
            'x_qe_xc_igcx':         7,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_PBE",
            'XC_functional_weight': 0.75,
            'exx_compute_weight': lambda exx: (1.0 - exx),
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
            'XC_functional_weight': 0.75,
            'exx_compute_weight': lambda exx: (1.0 - exx),
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "pb0x",
            'x_qe_xc_igcx_comment':    "PBE0 (PBE exchange*0.75)",
            'x_qe_xc_igcx':      8,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_B88",
            'XC_functional_weight': 0.72,
            'exx_compute_weight': lambda exx: 0.72 if abs(exx) > 0.01 else 1.0
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
            'XC_functional_weight': 0.8,
            'exx_compute_weight': lambda exx: (1.0 - exx),
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "b3lp",
            'x_qe_xc_igcx_comment':    "B3LYP (Becke88*0.72)",
            'x_qe_xc_igcx':      9,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_PBE_SOL",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "psx",
            'x_qe_xc_igcx_comment':    "PBEsol exchange",
            'x_qe_xc_igcx':      10,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_WC",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "wcx",
            'x_qe_xc_igcx_comment':    "Wu-Cohen",
            'x_qe_xc_igcx':      11,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "HYB_GGA_XC_HSE06",
            'exx_compute_weight': lambda exx: 1.0 if (abs(exx) > 0.01) else 0.0
        }, {
            'XC_functional_name': "GGA_X_PBE",
            'exx_compute_weight': lambda exx: 0.0 if (abs(exx) > 0.01) else 1.0
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }, {
            'XC_functional_name': 'GGA_C_PBE',
            'exx_compute_weight': lambda exx: 1.0 if (abs(exx) > 0.01) else 0.0
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "hse",
            'x_qe_xc_igcx_comment':    "HSE screened exchange",
            'x_qe_xc_igcx':      12,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_RPW86",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "rw86",
            'x_qe_xc_igcx_comment':    "revised PW86",
            'x_qe_xc_igcx':      13,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_PBE",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "pbe",
            'x_qe_xc_igcx_comment':    "same as PBX, back-comp.",
            'x_qe_xc_igcx':      14,
        },
    },
    {
        # igcx=15 is not defined in 5.4's funct.f90
        #        definition taken from 5.0, which did not have separate imeta
        'xc_terms': [{
            'XC_functional_name': "MGGA_X_TB09",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':    "tb09",
            'x_qe_xc_igcx_comment': "TB09 Meta-GGA (Espresso-version < 5.1)",
            'x_qe_xc_igcx':         15,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_C09X",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "c09x",
            'x_qe_xc_igcx_comment':    "Cooper 09",
            'x_qe_xc_igcx':      16,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_SOGGA",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "sox",
            'x_qe_xc_igcx_comment':    "sogga",
            'x_qe_xc_igcx':      17,
        },
    },
    {
        # igcx=18 is not defined in 5.4's funct.f90
        #        definition taken from 5.0, which did not have separate imeta
        'xc_terms': [{
            'XC_functional_name': "MGGA_X_M06_L",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':    "m6lx",
            'x_qe_xc_igcx_comment': "M06L Meta-GGA (Espresso-version < 5.1)",
            'x_qe_xc_igcx':         18,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_Q2D",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "q2dx",
            'x_qe_xc_igcx_comment':    "Q2D exchange grad corr",
            'x_qe_xc_igcx':      19,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "HYB_GGA_XC_GAU_PBE",
            'exx_compute_weight': lambda exx: 1.0 if (abs(exx) > 0.01) else 0.0
        }, {
            'XC_functional_name': "GGA_X_PBE",
            'exx_compute_weight': lambda exx: 0.0 if (abs(exx) > 0.01) else 1.0
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }, {
            'XC_functional_name': 'GGA_C_PBE',
            'exx_compute_weight': lambda exx: 1.0 if (abs(exx) > 0.01) else 0.0
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "gaup",
            'x_qe_xc_igcx_comment':    "Gau-PBE hybrid exchange",
            'x_qe_xc_igcx':      20,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_PW86",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "pw86",
            'x_qe_xc_igcx_comment':    "Perdew-Wang (1986) exchange",
            'x_qe_xc_igcx':      21,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_B86_MGC",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "b86b",
            'x_qe_xc_igcx_comment':    "Becke (1986) exchange",
            'x_qe_xc_igcx':      22,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_OPTB88_VDW",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "obk8",
            'x_qe_xc_igcx_comment':    "optB88  exchange",
            'x_qe_xc_igcx':      23,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_OPTB86_VDW",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "ob86",
            'x_qe_xc_igcx_comment':    "optB86b exchange",
            'x_qe_xc_igcx':      24,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_EV93",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "evx",
            'x_qe_xc_igcx_comment':    "Engel-Vosko exchange",
            'x_qe_xc_igcx':      25,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_B86_R",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "b86r",
            'x_qe_xc_igcx_comment':    "revised Becke (b86b)",
            'x_qe_xc_igcx':      26,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_LV_RPW86",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':       "cx13",
            'x_qe_xc_igcx_comment':    "consistent exchange",
            'x_qe_xc_igcx':      27,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_B88",
            'XC_functional_weight': 0.542,
            'exx_compute_weight':
                lambda exx: 0.542 if (abs(exx) > 0.01) else 1.0
        }, {
            'XC_functional_name': "GGA_X_PW91",
            'XC_functional_weight': 0.167,
            'exx_compute_weight':
                lambda exx: 0.167 if (abs(exx) > 0.01) else 0.0
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
            'exx_compute_weight':
                lambda exx: 0.709 if (abs(exx) > 0.01) else 1.0
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':    "x3lp",
            'x_qe_xc_igcx_comment': "X3LYP (Becke88*0.542 "
                                    " + Perdew-Wang91*0.167)",
            'x_qe_xc_igcx':         28,
        },
    },
# New additions for qe-v6.5.0:  igcx=[29..42]
# - - - - - -
# igcx: 29. The ingredient 'vdW-DF-cx' is documented in the nomad-meta-info, where it
# has the name 'vdw_c_df_cx'
# 'https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/-/wikis/metainfo/XC-functional':
# 'vdW-DF-cx' implies igcx=27 => 'cx13', hence LDA_X is implicit. Full weight.
    {
        'xc_terms': [{
            'XC_functional_name': "vdw_c_df_cx",
        }, {
        'XC_functional_name': 'HF_X',
        'exx_compute_weight': lambda exx: exx,
        'XC_functional_weight': 0.25,
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':    "cx0",
            'x_qe_xc_igcx_comment': "vdW-DF-cx+HF/4 (cx13-0)",
            'x_qe_xc_igcx':         29,
        },
    },
# - - - - - -
# igcx:30. Needs full LDA_X removal, due to 'GGA_X_RPW86' (see igcx:27)
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_RPW86",
        },{
            'XC_functional_name': 'HF_X',
            'exx_compute_weight': lambda exx: exx,
            'XC_functional_weight': 0.25,
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "r860",
            'x_qe_xc_igcx_comment': "rPW86+HF/4 (rw86-0); (for DF0)",
            'x_qe_xc_igcx': 30,
        },
    },
# - - - - - -
# igcx:31. Similar comments as in 'igcx:29'
    {
        'xc_terms': [{
            'XC_functional_name': "vdw_c_df_cx",
        }, {
            'XC_functional_name': 'HF_X',
            'exx_compute_weight': lambda exx: exx,
            'XC_functional_weight': 0.20,
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name':    "cx0p",
            'x_qe_xc_igcx_comment': "vdW-DF-cx+HF/5 (cx13-0p)",
            'x_qe_xc_igcx':         31,
        },
    },
# - - - - - -
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_RESERVED",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "ahcx",
            'x_qe_xc_igcx_comment': "vdW-DF-cx based; not yet in use (reserved PH)",
            'x_qe_xc_igcx': 32,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_RESERVED",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "ahf2",
            'x_qe_xc_igcx_comment': "vdW-DF2 based; not yet in use (reserved PH)",
            'x_qe_xc_igcx': 33,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_RESERVED",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "ahpb",
            'x_qe_xc_igcx_comment': "PBE based; not yet in use (reserved PH)",
            'x_qe_xc_igcx': 34,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_RESERVED",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "ahps",
            'x_qe_xc_igcx_comment': "PBE-sol based; not in use (reserved PH)",
            'x_qe_xc_igcx': 35,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_RESERVED",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "cx14",
            'x_qe_xc_igcx_comment': "Exporations (typo?: explorations), (reserved PH)",
            'x_qe_xc_igcx': 36,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_RESERVED",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "cx15",
            'x_qe_xc_igcx_comment': "Exporations (typo? explorations?)(reserved PH)",
            'x_qe_xc_igcx': 37,
        },
    },
#
# igcx': 38. Ingredients:
#   'b86r' -> 'igcx:26' -> 'GGA_X_B86_R'
#   'vdW-DF2' -> 'vdw_c_df2' . See nomad's gitlab:
#   'https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/-/wikis/metainfo/XC-functional':

    {
     'xc_terms': [{
            'XC_functional_name': "vdw_c_df2",
        }, {
            'XC_functional_name': "GGA_X_B86_R",
        }, {
            'XC_functional_name': 'HF_X',
            'exx_compute_weight': lambda exx: exx,
            'XC_functional_weight': 0.25,
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "br0",
            'x_qe_xc_igcx_comment': "vdW-DF2-b86r+HF/4 (b86r-0)",
            'x_qe_xc_igcx': 38,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_RESERVED",
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "cx16",
            'x_qe_xc_igcx_comment': "Exporations (typo?, explorations?)(reserved PH)",
            'x_qe_xc_igcx': 39,
        },
    },
#- - - -
    {
        'xc_terms': [{
            'XC_functional_name': "vdw_c_df1",
        }, {
            'XC_functional_name': "GGA_X_C09X",
        }, {
            'XC_functional_name': 'HF_X',
            'exx_compute_weight': lambda exx: exx,
            'XC_functional_weight': 0.25,
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "c090",
            'x_qe_xc_igcx_comment': "vdW-DF-c09+HF/4 (c09-0)",
            'x_qe_xc_igcx': 40,
        },
    },
# - - - - - - -
# 'igcx:41' Note: 'B86b' is defined in 'igcx:22'
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_B86_MGC",
        }],
            'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
            'XC_functional_weight': 0.75,
            'exx_compute_weight': lambda exx: (1.0 - exx),
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "b86x",
            'x_qe_xc_igcx_comment': "B86b exchange * 0.75",
            'x_qe_xc_igcx': 41,
        },
    },
# - - - - - - -
# 'B88' is defined in 'igcx:1'
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_X_B88",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
            'XC_functional_weight': 0.50,
            'exx_compute_weight': lambda exx: (1.0 - exx),
        }],
        'xc_section_method': {
            'x_qe_xc_igcx_name': "b88x",
            'x_qe_xc_igcx_comment': "B88 exchange * 0.50",
            'x_qe_xc_igcx': 42,
        },
    },
]

# UNchanged between espresso v5.4 & v6.5
CORRELATION_GRADIENT_CORRECTION = [
    None,
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_C_P86",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "p86",
            'x_qe_xc_igcc_comment':    "Perdew86",
            'x_qe_xc_igcc':      1,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_C_PW91",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "ggc",
            'x_qe_xc_igcc_comment':    "Perdew-Wang 91 corr.",
            'x_qe_xc_igcc':      2,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_C_LYP",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_LYP',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "blyp",
            'x_qe_xc_igcc_comment':    "Lee-Yang-Parr",
            'x_qe_xc_igcc':      3,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_C_PBE",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "pbc",
            'x_qe_xc_igcc_comment':    "Perdew-Burke-Ernzenhof corr",
            'x_qe_xc_igcc':      4,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_XC_HCTH_120",
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "hcth",
            'x_qe_xc_igcc_comment':    "Cambridge exch, Handy et al, HCTH/120",
            'x_qe_xc_igcc':      5,
        },
    },
    {
        # igcc=6 is not defined in 5.4's funct.f90
        #        definition taken from 5.0, which did not have separate imeta
        'xc_terms': [{
            'XC_functional_name': "MGGA_C_TPSS",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':    "tpss",
            'x_qe_xc_igcc_comment': "TPSS Meta-GGA (Espresso-version < 5.1)",
            'x_qe_xc_igcc':         6,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_C_LYP",
            'XC_functional_weight': 0.81,
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_LYP',
            'XC_functional_weight': 0.81,
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "b3lp",
            'x_qe_xc_igcc_comment':    "B3LYP (Lee-Yang-Parr*0.81)",
            'x_qe_xc_igcc':      7,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_C_PBE_SOL",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "psc",
            'x_qe_xc_igcc_comment':    "PBEsol corr",
            'x_qe_xc_igcc':      8,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_C_PBE",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "pbe",
            'x_qe_xc_igcc_comment':    "same as PBX, back-comp.",
            'x_qe_xc_igcc':      9,
        },
    },
    {
        # igcc=10 is not defined in 5.4's funct.f90
        #        definition taken from 5.0, which did not have separate imeta
        #        functionals.f90 tells that correlation is taken from tpss
        'xc_terms': [{
            'XC_functional_name': "MGGA_C_TPSS",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':    "tb09",
            'x_qe_xc_igcc_comment': "TB09 Meta-GGA (Espresso-version < 5.1)",
            'x_qe_xc_igcc':         10,
        },
    },
    {
        # igcc=11 is not defined in 5.4's funct.f90
        #        definition taken from 5.0, which did not have separate imeta
        'xc_terms': [{
            'XC_functional_name': "MGGA_C_M06_L",
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':    "m6lc",
            'x_qe_xc_igcc_comment': "M06L Meta-GGA (Espresso-version < 5.1)",
            'x_qe_xc_igcc':         11,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_C_Q2D",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "q2dc",
            'x_qe_xc_igcc_comment':    "Q2D correlation grad corr",
            'x_qe_xc_igcc':      12,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "GGA_C_LYP",
            'XC_functional_weight': 0.871,
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_LYP',
            'XC_functional_weight': 0.871,
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':       "x3lp",
            'x_qe_xc_igcc_comment':    "X3LYP (Lee-Yang-Parr*0.871)",
            'x_qe_xc_igcc':      13,
        },
    },
    {
    #  igcc=14 is not defined in NEITHER of v5.1, v6.1, v6.4's Modules/funct.f90
    # "BEEF-vdW, a GGA with vdW-DF2 type nonlocal correlation"
        'xc_terms': [{
            'XC_functional_name':  "GGA_C_BEEF-vdW",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_igcc_name':  "BEEF-vdW",
            'x_qe_xc_igcc_comment':  "libbeef V0.1.1 library",
            'x_qe_xc_igcc':  14,
        },
    },
]

# New additions for espresso-6.5.0:  imeta=[4, 5, 6]
META_GGA = [
    None,
    {
        'xc_terms': [{
            'XC_functional_name': "MGGA_X_TPSS",
        }, {
            'XC_functional_name': "MGGA_C_TPSS",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }, {
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_imeta_name':       "tpss",
            'x_qe_xc_imeta_comment':    "TPSS Meta-GGA",
            'x_qe_xc_imeta':      1,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "MGGA_X_M06_L",
        }, {
            'XC_functional_name': "MGGA_C_M06_L",
        }],
        'xc_section_method': {
            'x_qe_xc_imeta_name':       "m6lx",
            'x_qe_xc_imeta_comment':    "M06L Meta-GGA",
            'x_qe_xc_imeta':      2,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "MGGA_X_TB09",
        }, {
            # confirmed by looking into functionals.f90
            'XC_functional_name': "MGGA_C_TPSS",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }, {
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_imeta_name':       "tb09",
            'x_qe_xc_imeta_comment':    "TB09 Meta-GGA",
            'x_qe_xc_imeta':      3,
        },
    },
    # imeta = [4,5,6] are new espresso-6.5.0/Modules/funct.f90
    {
        'xc_terms': [{
            'XC_functional_name': "MGGA_X_TPSS",
        }, {
            'XC_functional_name': "MGGA_C_TPSS",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }, {
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_imeta_name': "+meta",
            'x_qe_xc_imeta_comment': "activate MGGA even without MGGA-XC",
            'x_qe_xc_imeta': 4,
        },
    },
    # - - - - - - - -
    {
        'xc_terms': [{
            'XC_functional_name': "MGGA_X_SCAN",
        }, {
            'XC_functional_name': "MGGA_C_SCAN",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }, {
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_imeta_name': "scan",
            'x_qe_xc_imeta_comment': "SCAN Meta-GGA ",
            'x_qe_xc_imeta': 5,
        },
    },
# - - - - - - - -
    {
        'xc_terms': [{
            'XC_functional_name': "HYB_MGGA_X_SCAN0",
        }, {
            'XC_functional_name': "MGGA_C_SCAN",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': 'LDA_X',
        }, {
            'XC_functional_name': 'LDA_C_PW',
        }],
        'xc_section_method': {
            'x_qe_xc_imeta_name': "sca0",
            'x_qe_xc_imeta_comment': "SCAN0  Meta-GGA",
            'x_qe_xc_imeta': 6,
        },
    },
]

VAN_DER_WAALS = [
    None,
    {
        'xc_terms': [{
            'XC_functional_name': "VDW_C_DF1",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': "LDA_C_PW",
        }],
        'xc_section_method': {
            'x_qe_xc_inlc_name':       "vdw1",
            'x_qe_xc_inlc_comment':    "vdW-DF1",
            'x_qe_xc_inlc':      1,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "VDW_C_DF2",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': "LDA_C_PW",
        }],
        'xc_section_method': {
            'x_qe_xc_inlc_name':       "vdw2",
            'x_qe_xc_inlc_comment':    "vdW-DF2",
            'x_qe_xc_inlc':      2,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "VDW_C_RVV10",
        }],
        'xc_terms_remove': [{
            'XC_functional_name': "GGA_C_PBE",
        }],
        'xc_section_method': {
            'x_qe_xc_inlc_name':       "vv10",
            'x_qe_xc_inlc_comment':    "rVV10",
            'x_qe_xc_inlc':      3,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "VDW_DFX_x_qe",
        }],
        'xc_section_method': {
            'x_qe_xc_inlc_name':    "vdwx",
            'x_qe_xc_inlc_comment': "vdW-DF-x (reserved Thonhauser,"
                                    " not implemented)",
            'x_qe_xc_inlc':         4,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "VDW_DFY_x_qe",
        }],
        'xc_section_method': {
            'x_qe_xc_inlc_name':    "vdwy",
            'x_qe_xc_inlc_comment': "vdW-DF-y (reserved Thonhauser,"
                                    " not implemented)",
            'x_qe_xc_inlc':         5,
        },
    },
    {
        'xc_terms': [{
            'XC_functional_name': "VDW_DFZ_x_qe",
        }],
        'xc_section_method': {
            'x_qe_xc_inlc_name':    "vdwz",
            'x_qe_xc_inlc_comment': "vdW-DF-z (reserved Thonhauser,"
                                    " not implemented)",
            'x_qe_xc_inlc':         6,
        },
    },
]


XC_COMPONENT = [
    EXCHANGE,
    CORRELATION,
    EXCHANGE_GRADIENT_CORRECTION,
    CORRELATION_GRADIENT_CORRECTION,
    VAN_DER_WAALS,
    META_GGA,
]


XC_COMPONENT_NAME = [
    'EXCHANGE',
    'CORRELATION',
    'EXCHANGE_GRADIENT_CORRECTION',
    'CORRELATION_GRADIENT_CORRECTION',
    'VAN_DER_WAALS',
    'META_GGA',
]

LIBXC_SHORTCUT = {
    "0.810*GGA_C_LYP+0.720*GGA_X_B88+0.200*HF_X+0.190*LDA_C_VWN": {
        'xc_terms': [{
            'XC_functional_name': "HYB_GGA_XC_B3LYP",
        }]
    },
    "GGA_C_PBE+0.750*GGA_X_PBE+0.250*HF_X": {
        'xc_terms': [{
            'XC_functional_name': "HYB_GGA_XC_PBEH",
        }]
    },
}



# Copyright 2016-2018 The NOMAD Developers Group
# Copyright 2017-2018 Lorenzo Pardini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import xml.sax
import logging
import numpy as np
from nomadcore.unit_conversion.unit_conversion import convert_unit_function
from nomadcore.unit_conversion.unit_conversion import convert_unit
from nomadcore.unit_conversion import unit_conversion


class DosHandler(xml.sax.handler.ContentHandler):
    def __init__(self, backend, spinTreat, unitCellVol, fermiEnergy):
        self.backend = backend
        self.dosSectionGIndex = -1
        self.inDos = False
        self.dosProjSectionGIndex = -1
        self.inDosProj = False
        self.spinTreat = spinTreat
        self.totDos = []
        self.totDosSpin = [[],[]]
        self.dosProj = []
        self.dosProjSpin = []
        self.energy = []
        self.energySpin = []
        self.speciesrn = []
        self.atom = []
        self.numDosVal = 0
        self.dosProjDummy = []
        self.dosProjDummy2 = []
        self.unitCellVol = float(unitCellVol[0])
        self.fermiEnergy = fermiEnergy

    def endDocument(self):
        self.inDosProj = False
        self.backend.closeSection("section_dos",self.dosSectionGIndex)
        self.dosSectionGIndex = -1

    def startElement(self, name, attrs):
        ha_per_joule = convert_unit(1, "hartree", "J")
        joule_in_ev = convert_unit(1, "eV", "J")
        fromH = unit_conversion.convert_unit_function("hartree", "J")
        if name == "totaldos":
            self.dosSectionGIndex = self.backend.openSection("section_dos")
            self.inDos = True
        elif name == "partialdos":
            self.speciesrn.append(int(attrs.getValue('speciesrn')))
            self.atom.append(int(attrs.getValue('atom')))
            self.dosProjDummy.append([])
            self.dosProjDummy2.append([])
            if self.speciesrn [-1] == 1 and self.atom[-1] == 1:
                self.dosProjSectionGIndex = self.backend.openSection("section_atom_projected_dos")
                self.inDosProj = True
        elif name == "point":
            if self.inDos:
                self.totDos.append(float(attrs.getValue('dos'))/ha_per_joule)
                self.energy.append(fromH(float(attrs.getValue('e'))))
            elif self.inDosProj:
                self.dosProj.append(float(attrs.getValue('dos'))/ha_per_joule)
                self.energy.append(fromH(float(attrs.getValue('e'))))
        elif name == "diagram":
            if not self.speciesrn: pass
            elif self.speciesrn [-1] == 1 and self.atom[-1] == 1:
                if not self.spinTreat:
                    self.dosProjSpin.append([])
                else:
                    nspin = int(attrs.getValue("nspin"))
                    if nspin == 1:
                        self.dosProjSpin.append([])
        elif name == "limrep":
            for i in range(0,len(self.dosProjSpin)):
                for j in range(0,2):
                    self.dosProjSpin[i].append([])
                    for k in range(0,len(self.speciesrn)):
                        self.dosProjDummy2[k].append([])
                        self.dosProjSpin[i][j].append([])
            for i in range (0,len(self.speciesrn)):
                if not self.spinTreat:
                    self.dosProjDummy[i] = self.dosProj[i*len(self.dosProjSpin)*self.numDosVal:(i+1)*len(self.dosProjSpin)*self.numDosVal]
                else:
                    self.dosProjDummy[i] = self.dosProj[i*int(2*len(self.dosProjSpin))*self.numDosVal:(i+1)*int(2*len(self.dosProjSpin))*self.numDosVal]
            for j in range(0,int(2*len(self.dosProjSpin))):
                for i in range (0,len(self.speciesrn)):
                    self.dosProjDummy2[i][j] = self.dosProjDummy[i][j*self.numDosVal:(j+1)*self.numDosVal]
            if not self.spinTreat:
                for i in range(0,len(self.dosProjSpin)):
                    for j in range(0,len(self.speciesrn)):
                        self.dosProjSpin[i][0][j] = self.dosProjDummy2[j][i]
                        self.dosProjSpin[i][1][j] = self.dosProjDummy2[j][i]
            else:
                for j in range(0,len(self.speciesrn)):
                    for i in range(0,int(2*len(self.dosProjSpin))):
                        if i < len(self.dosProjSpin):
                            self.dosProjSpin[i][0][j] = self.dosProjDummy2[j][i]
                        else:
                            k = int(i - len(self.dosProjSpin))
                            self.dosProjSpin[k][1][j] = self.dosProjDummy2[j][i]
    def endElement(self, name):
        if name == 'totaldos':
            self.inDos = False
            if not self.spinTreat:
                self.numDosVal = len(self.energy)
                self.totDosSpin[0] = self.totDos[0:self.numDosVal]
                self.totDosSpin[1] = self.totDos[0:self.numDosVal]
                self.energySpin = self.energy[0:self.numDosVal]
                self.backend.addValue("dos_values", np.array(self.totDosSpin) * self.unitCellVol)
                self.backend.addValue("dos_energies", self.energySpin + self.fermiEnergy)
                self.backend.addValue("number_of_dos_values", self.numDosVal)
                self.backend.addValue("dos_kind", "electronic")
            else:
                self.numDosVal = int(len(self.energy) / 2)
                self.totDosSpin[0] = self.totDos[0:self.numDosVal]
                self.totDosSpin[1] = self.totDos[self.numDosVal:int(2 * (self.numDosVal))]
                self.energySpin = self.energy[0:self.numDosVal]
                self.backend.addValue("dos_values", np.array(self.totDosSpin) * self.unitCellVol)
                self.backend.addValue("dos_energies", self.energySpin + self.fermiEnergy)
                self.backend.addValue("number_of_dos_values", self.numDosVal)

        elif name == 'partialdos':
            pass
        elif name == 'interstitialdos':
            self.backend.addValue("atom_projected_dos_values_lm",self.dosProjSpin)
            self.backend.addValue("number_of_lm_atom_projected_dos",len(self.dosProjSpin))
            self.backend.addValue("number_of_atom_projected_dos_values",self.numDosVal)
            self.backend.addValue("atom_projected_dos_energies",self.energy[0:self.numDosVal])
            self.backend.addValue("atom_projected_dos_m_kind","spherical")
            self.backend.closeSection("section_atom_projected_dos",self.dosProjSectionGIndex)
            self.dosProjSectionGIndex = -1
            self.inDosProj = False
    def startElementNS(self, name, qname, attrs):
        attrDict={}
        for name in attrs.getNames():
            attrDict[name] = attrs.getValue(name)
        logging.error("start element %s ns %s attr %s", name, qname, attrDict)

    def endElementNS(self, name, qname):
        logging.error("end element %s ns %s", name, qname)

    def characters(self, content):
        pass

def parseDos(inF, backend, spinTreat, unitCellVol, fermiEnergy):
    handler = DosHandler(backend, spinTreat, unitCellVol, fermiEnergy)
    logging.info("will parse")
    xml.sax.parse(inF, handler)
    logging.info("did parse")



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='elk_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='elk.nomadmetainfo.json'))


class x_elk_section_lattice_vectors(MSection):
    '''
    lattice vectors
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_elk_section_lattice_vectors'))

    x_elk_geometry_lattice_vector_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x component of lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_lattice_vector_x'))

    x_elk_geometry_lattice_vector_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y component of lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_lattice_vector_y'))

    x_elk_geometry_lattice_vector_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z component of lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_lattice_vector_z'))


class x_elk_section_reciprocal_lattice_vectors(MSection):
    '''
    reciprocal lattice vectors
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_elk_section_reciprocal_lattice_vectors'))

    x_elk_geometry_reciprocal_lattice_vector_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter',
        description='''
        x component of reciprocal lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_reciprocal_lattice_vector_x'))

    x_elk_geometry_reciprocal_lattice_vector_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter',
        description='''
        y component of reciprocal lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_reciprocal_lattice_vector_y'))

    x_elk_geometry_reciprocal_lattice_vector_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter',
        description='''
        z component of reciprocal lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_reciprocal_lattice_vector_z'))


class x_elk_section_atoms_group(MSection):
    '''
    a group of atoms of the same type
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_elk_section_atoms_group'))

    x_elk_geometry_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        labels of atom
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_atom_labels'))

    x_elk_geometry_atom_number = Quantity(
        type=str,
        shape=[],
        description='''
        number to identify the atoms of a species
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_atom_number'))

    x_elk_geometry_atom_positions_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x component of atomic position
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_atom_positions_x'))

    x_elk_geometry_atom_positions_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y component of atomic position
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_atom_positions_y'))

    x_elk_geometry_atom_positions_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z component of atomic position
        ''',
        a_legacy=LegacyDefinition(name='x_elk_geometry_atom_positions_z'))


class x_elk_section_spin(MSection):
    '''
    section for exciting spin treatment
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_elk_section_spin'))

    x_elk_spin_treatment = Quantity(
        type=str,
        shape=[],
        description='''
        Spin treatment
        ''',
        a_legacy=LegacyDefinition(name='x_elk_spin_treatment'))


class x_elk_section_xc(MSection):
    '''
    index for elk functional
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_elk_section_xc'))

    x_elk_xc_functional = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        index for elk functional
        ''',
        a_legacy=LegacyDefinition(name='x_elk_xc_functional'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_elk_brillouin_zone_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter ** 3',
        description='''
        Brillouin zone volume
        ''',
        a_legacy=LegacyDefinition(name='x_elk_brillouin_zone_volume'))

    x_elk_simulation_reciprocal_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='meter',
        description='''
        Reciprocal lattice vectors of the simulation cell.
        ''',
        categories=[public.configuration_core],
        a_legacy=LegacyDefinition(name='x_elk_simulation_reciprocal_cell'))

    x_elk_unit_cell_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter ** 3',
        description='''
        unit cell volume
        ''',
        a_legacy=LegacyDefinition(name='x_elk_unit_cell_volume'))

    x_elk_muffin_tin_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        muffin-tin radius
        ''',
        a_legacy=LegacyDefinition(name='x_elk_muffin_tin_radius'))

    x_elk_muffin_tin_points = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        muffin-tin points
        ''',
        a_legacy=LegacyDefinition(name='x_elk_muffin_tin_points'))

    x_elk_number_kpoint_x = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number k-points x
        ''',
        a_legacy=LegacyDefinition(name='x_elk_number_kpoint_x'))

    x_elk_number_kpoint_y = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number k-points y
        ''',
        a_legacy=LegacyDefinition(name='x_elk_number_kpoint_y'))

    x_elk_number_kpoint_z = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number k-points z
        ''',
        a_legacy=LegacyDefinition(name='x_elk_number_kpoint_z'))

    x_elk_number_kpoints = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number k-points
        ''',
        a_legacy=LegacyDefinition(name='x_elk_number_kpoints'))

    x_elk_kpoint_offset_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        K-points offset x component
        ''',
        a_legacy=LegacyDefinition(name='x_elk_kpoint_offset_x'))

    x_elk_kpoint_offset_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        K-points offset y component
        ''',
        a_legacy=LegacyDefinition(name='x_elk_kpoint_offset_y'))

    x_elk_kpoint_offset_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        K-points offset z component
        ''',
        a_legacy=LegacyDefinition(name='x_elk_kpoint_offset_z'))

    x_elk_rgkmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        Radius MT * Gmax
        ''',
        a_legacy=LegacyDefinition(name='x_elk_rgkmax'))

    x_elk_gvector_size_x = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        G-vector grid size x
        ''',
        a_legacy=LegacyDefinition(name='x_elk_gvector_size_x'))

    x_elk_gvector_size_y = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        G-vector grid size y
        ''',
        a_legacy=LegacyDefinition(name='x_elk_gvector_size_y'))

    x_elk_gvector_size_z = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        G-vector grid size z
        ''',
        a_legacy=LegacyDefinition(name='x_elk_gvector_size_z'))

    x_elk_gvector_total = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        G-vector total
        ''',
        a_legacy=LegacyDefinition(name='x_elk_gvector_total'))

    x_elk_lmaxapw = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Angular momentum cut-off for the APW functions
        ''',
        a_legacy=LegacyDefinition(name='x_elk_lmaxapw'))

    x_elk_gkmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter',
        description='''
        Maximum length of |G+k| for APW functions
        ''',
        a_legacy=LegacyDefinition(name='x_elk_gkmax'))

    x_elk_smearing_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Smearing width for KS occupancies
        ''',
        a_legacy=LegacyDefinition(name='x_elk_smearing_width'))

    x_elk_lo = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Total number of local-orbitals
        ''',
        a_legacy=LegacyDefinition(name='x_elk_lo'))

    x_elk_gmaxvr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter',
        description='''
        Maximum length of |G|
        ''',
        a_legacy=LegacyDefinition(name='x_elk_gmaxvr'))

    x_elk_valence_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Total number of valence states
        ''',
        a_legacy=LegacyDefinition(name='x_elk_valence_states'))

    x_elk_core_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Total number of core states
        ''',
        a_legacy=LegacyDefinition(name='x_elk_core_states'))

    x_elk_empty_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of empty states
        ''',
        a_legacy=LegacyDefinition(name='x_elk_empty_states'))

    x_elk_wigner_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        Effective Wigner radius
        ''',
        a_legacy=LegacyDefinition(name='x_elk_wigner_radius'))

    x_elk_electronic_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Electronic charge
        ''',
        a_legacy=LegacyDefinition(name='x_elk_electronic_charge'))

    x_elk_valence_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Valence charge
        ''',
        a_legacy=LegacyDefinition(name='x_elk_valence_charge'))

    x_elk_nuclear_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Nuclear charge
        ''',
        a_legacy=LegacyDefinition(name='x_elk_nuclear_charge'))

    x_elk_core_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Core charge
        ''',
        a_legacy=LegacyDefinition(name='x_elk_core_charge'))

    x_elk_section_lattice_vectors = SubSection(
        sub_section=SectionProxy('x_elk_section_lattice_vectors'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_elk_section_lattice_vectors'))

    x_elk_section_reciprocal_lattice_vectors = SubSection(
        sub_section=SectionProxy('x_elk_section_reciprocal_lattice_vectors'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_elk_section_reciprocal_lattice_vectors'))

    x_elk_section_atoms_group = SubSection(
        sub_section=SectionProxy('x_elk_section_atoms_group'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_elk_section_atoms_group'))

    x_elk_section_spin = SubSection(
        sub_section=SectionProxy('x_elk_section_spin'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_elk_section_spin'))

    x_elk_section_xc = SubSection(
        sub_section=SectionProxy('x_elk_section_xc'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_elk_section_xc'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    x_elk_core_charge_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Core charge scf iteration
        ''',
        a_legacy=LegacyDefinition(name='x_elk_core_charge_scf_iteration'))

    x_elk_valence_charge_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Valence charge scf iteration
        ''',
        a_legacy=LegacyDefinition(name='x_elk_valence_charge_scf_iteration'))

    x_elk_interstitial_charge_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Interstitial charge scf iteration
        ''',
        a_legacy=LegacyDefinition(name='x_elk_interstitial_charge_scf_iteration'))

    x_elk_fermi_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Fermi energy
        ''',
        a_legacy=LegacyDefinition(name='x_elk_fermi_energy_scf_iteration'))

    x_elk_core_electron_kinetic_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Core-electron kinetic energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_core_electron_kinetic_energy_scf_iteration'))

    x_elk_coulomb_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Coulomb energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_coulomb_energy_scf_iteration'))

    x_elk_coulomb_potential_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Coulomb potential energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_coulomb_potential_energy_scf_iteration'))

    x_elk_nuclear_nuclear_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Nuclear-nuclear energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_nuclear_nuclear_energy_scf_iteration'))

    x_elk_electron_nuclear_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Electron-nuclear energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_electron_nuclear_energy_scf_iteration'))

    x_elk_hartree_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Hartree energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_hartree_energy_scf_iteration'))

    x_elk_madelung_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Madelung energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_madelung_energy_scf_iteration'))

    x_elk_exchange_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Exchange energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_exchange_energy_scf_iteration'))

    x_elk_correlation_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Correlation energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_correlation_energy_scf_iteration'))

    x_elk_electron_entropic_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Electron entropic energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_electron_entropic_energy_scf_iteration'))

    x_elk_dos_fermi_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / joule',
        description='''
        DOS at Fermi energy
        ''',
        a_legacy=LegacyDefinition(name='x_elk_dos_fermi_scf_iteration'))

    x_elk_direct_gap_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Estimated fundamental direct gap
        ''',
        a_legacy=LegacyDefinition(name='x_elk_direct_gap_scf_iteration'))

    x_elk_indirect_gap_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Estimated fundamental indirect gap
        ''',
        a_legacy=LegacyDefinition(name='x_elk_indirect_gap_scf_iteration'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_elk_core_charge_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Core charge final
        ''',
        a_legacy=LegacyDefinition(name='x_elk_core_charge_final'))

    x_elk_valence_charge_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Valence charge final
        ''',
        a_legacy=LegacyDefinition(name='x_elk_valence_charge_final'))

    x_elk_interstitial_charge_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Interstitial charge final
        ''',
        a_legacy=LegacyDefinition(name='x_elk_interstitial_charge_final'))

    x_elk_fermi_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Fermi energy final
        ''',
        a_legacy=LegacyDefinition(name='x_elk_fermi_energy'))

    x_elk_core_electron_kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Core-electron kinetic energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_core_electron_kinetic_energy'))

    x_elk_coulomb_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Coulomb energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_coulomb_energy'))

    x_elk_coulomb_potential_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Coulomb potential energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_coulomb_potential_energy'))

    x_elk_nuclear_nuclear_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Nuclear-nuclear energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_nuclear_nuclear_energy'))

    x_elk_electron_nuclear_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Electron-nuclear energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_electron_nuclear_energy'))

    x_elk_hartree_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Hartree energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_hartree_energy'))

    x_elk_madelung_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Madelung energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_madelung_energy'))

    x_elk_exchange_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Exchange energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_exchange_energy'))

    x_elk_correlation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Correlation energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_correlation_energy'))

    x_elk_electron_entropic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Electron entropic energy final
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_elk_electron_entropic_energy'))

    x_elk_dos_fermi = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / joule',
        description='''
        DOS at Fermi energy
        ''',
        a_legacy=LegacyDefinition(name='x_elk_dos_fermi'))

    x_elk_direct_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Estimated fundamental direct gap final
        ''',
        a_legacy=LegacyDefinition(name='x_elk_direct_gap'))

    x_elk_indirect_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Estimated fundamental indirect gap final
        ''',
        a_legacy=LegacyDefinition(name='x_elk_indirect_gap'))


m_package.__init_metainfo__()



# Copyright 2018-2018 Berk Onat, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import map
from builtins import range
from builtins import object
import numpy as np
import nomadcore.ActivateLogging
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.smart_parser import SmartParserCommon as SmartParser
from nomadcore.smart_parser.SmartParserCommon import get_metaInfo, conv_str, conv_int, conv_float, open_section
from nomadcore.smart_parser.SmartParserDictionary import getList_MetaStrInDict, getDict_MetaStrInDict
from nomadcore.smart_parser.SmartParserDictionary import isMetaStrInDict
from .NAMDDictionary import get_updateDictionary, set_Dictionaries
from .NAMDCommon import PARSERNAME, PROGRAMNAME, PARSERVERSION, PARSERTAG, LOGGER
from .NAMDCommon import PARSER_INFO_DEFAULT, META_INFO_PATH, set_excludeList, set_includeList
from nomadcore.md_data_access import MDDataAccess as MDDA
import argparse
import logging
import os
import re
import sys
import datetime
import io
from nomadcore.simple_parser import mainFunction

############################################################
# This is the parser for the main file of NAMD.
############################################################

#PRINTABLE = re.compile(r"\W+")

class NAMDParser(SmartParser.ParserBase):
    """Context for parsing NAMD main file.

    This class keeps tracks of several NAMD settings to adjust the parsing to them.
    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of writing and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """
    def __init__(self):
        # dictionary of energy values, which are tracked between SCF iterations and written after convergence
        self.totalEnergyList = {
                                'energy_electrostatic': None,
                                'energy_total_T0_per_atom': None,
                                'energy_free_per_atom': None,
                               }
        SmartParser.ParserBase.__init__(
            self, re_program_name=re.compile(r"\s*"+PROGRAMNAME+"$"),
            parsertag=PARSERTAG, metainfopath=META_INFO_PATH,
            parserinfodef=PARSER_INFO_DEFAULT)

        set_Dictionaries(self)
        self.metaInfoEnv = get_metaInfo(self)
        self.secGIndexDict = {}

        self.cachingLevelForMetaName = {
                               PARSERTAG + '_trajectory_file_detect': CachingLevel.Cache,
                               PARSERTAG + '_geometry_optimization_cdetect': CachingLevel.Cache,
                               PARSERTAG + '_mdin_finline': CachingLevel.Ignore,
                               #PARSERTAG + '_section_input_output_files': CachingLevel.Ignore,
                               PARSERTAG + '_single_configuration_calculation_detect': CachingLevel.Cache,
                              }
        for name in self.metaInfoEnv.infoKinds:
            metaInfo = self.metaInfoEnv.infoKinds[name]
            if (name.startswith(PARSERTAG + '_mdin_') and
                metaInfo.kindStr == "type_document_content" and
                (PARSERTAG + "_mdin_method" in metaInfo.superNames or
                 PARSERTAG + "_mdin_run" in metaInfo.superNames or
                 PARSERTAG + "_mdin_system" in metaInfo.superNames) or
                name.startswith(PARSERTAG + '_parm_') and
                metaInfo.kindStr == "type_document_content" and
                (PARSERTAG + "_mdin_method" in metaInfo.superNames or
                 PARSERTAG + "_mdin_run" in metaInfo.superNames or
                 PARSERTAG + "_mdin_system" in metaInfo.superNames) or
                #name.startswith(PARSERTAG + '_mdin_file_') and
                name.startswith(PARSERTAG + '_inout_file_') or
                #metaInfo.kindStr == "type_document_content" and
                #(PARSERTAG + "_section_input_output_files" in metaInfo.superNames or
                # "section_run" in metaInfo.superNames) or
                name.startswith(PARSERTAG + '_inout_control_') or
                #(PARSERTAG + "_section_control_parameters" in metaInfo.superNames) or
                #name.startswith(PARSERTAG + '_mdin_') and
                #(PARSERTAG + "_section_control_parameters" in metaInfo.superNames) or
                name.startswith(PARSERTAG + '_mdout_') or
                name.startswith(PARSERTAG + '_mdout_') and
                #metaInfo.kindStr == "type_document_content" and
                (PARSERTAG + "_mdout_method" in metaInfo.superNames or
                 PARSERTAG + "_mdout_system" in metaInfo.superNames or
                 "section_run" in metaInfo.superNames or
                 PARSERTAG + "_mdout_single_configuration_calculation" in metaInfo.superNames)
                or name.startswith('section_single_configuration_calculation')
                or PARSERTAG + '_mdout' in metaInfo.superNames
                or 'section_sampling_method' in metaInfo.superNames
                or 'section_single_configuration_calculation' in metaInfo.superNames
               ):
                self.cachingLevelForMetaName[name] = CachingLevel.Cache
            if name in self.extraDict.keys():
                self.cachingLevelForMetaName[name] = CachingLevel.Ignore


    def initialize_values(self):
        """Initializes the values of certain variables.

        This allows a consistent setting and resetting of the variables,
        when the parsing starts and when a section_run closes.
        """
        set_Dictionaries(self)
        self.secGIndexDict.clear()
        self.secMethodGIndex = None
        self.secSystemGIndex = None
        self.secTopologyGIndex = None
        self.secSamplingGIndex = None
        self.secSingleGIndex = None
        self.secVDWGIndex = None
        self.secAtomType = None
        self.inputMethodIndex = None
        self.inputControlIndex = None
        self.mainMethodIndex = None
        self.mainCalcIndex = None
        self.MD = True
        self.topoDict = None
        self.topologyTable = None
        self.topologyBonds = None
        self.topology = None
        self.topologyFormat = None
        self.topologyFile = None
        self.trajectory = None
        self.trajectoryFormat = None
        self.trajectoryFile = None
        self.readChunk = 300
        self.boxlengths = None
        self.latticevectors = None
        self.atompositions = None
        self.inputcoords = None
        self.inputpositions = None
        self.outputcoords = None
        self.outputpositions = None
        # start with -1 since zeroth iteration is the initialization
        self.MDiter = -1
        self.MDstep = 1
        self.MDcurrentstep = -1
        self.MDnextstep = 0
        self.singleConfCalcs = []
        self.minConverged = None
        self.parsedLogFile = False
        self.LogSuperContext = None
        self.forces_raw = []
        self.ffparams = []
        self.lastCalculationGIndex = None
        self.logFileName = None
        self.lastfInLine = None
        self.lastfInMatcher = None
        self.secOpen = open_section
        self.superP = self.parser.backend.superBackend
        self.MDData = MDDA.MDDataAccess()
        if self.recordList:
            self.recordList.close()
        self.recordList = io.StringIO()

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts.

        Get compiled parser, filename and metadata.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        self.fName = fInName
        # save metadata
        self.metaInfoEnv = self.parser.parserBuilder.metaInfoEnv
        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()

    #def peekline(self, parser):
    #    pos = parser.fIn.fIn.tell()
    #    line = parser.fIn.fIn.readline()
    #    parser.fIn.fIn.seek(pos)
    #    return line

    def onClose_section_run(self, backend, gIndex, section):
        """Trigger called when section_run is closed.

        Write the keywords from control parametres and
        the NAMD output from the parsed log output,
        which belong to settings_run.
        Variables are reset to ensure clean start for new run.
        """
        frameSequenceGIndex = backend.openSection("section_frame_sequence")
        section_frameseq_Dict = get_updateDictionary(self, 'frameseqend')
        updateFrameDict = {
                'startSection' : [
                    ['section_frame_sequence']],
                #'muteSections' : [['section_sampling_method']],
                'dictionary' : section_frameseq_Dict
                }
        self.metaStorage.update(updateFrameDict)
        self.metaStorage.updateBackend(backend.superBackend,
                startsection=['section_frame_sequence'],
                autoopenclose=False)
        backend.addValue("frame_sequence_to_sampling_ref", self.secSamplingGIndex)
        backend.addArrayValues("frame_sequence_local_frames_ref", np.asarray(self.singleConfCalcs))
        backend.closeSection("section_frame_sequence", frameSequenceGIndex)

        # reset all variables
        self.initialize_values()

    def onClose_x_namd_section_input_output_files(self, backend, gIndex, section):
        """Trigger called when x_namd_section_input_output_files is closed.

        Determine whether topology, trajectory and input coordinate files are
        supplied to the parser

        Initiates topology and trajectory file handles.

        Captures topology, atomic positions, atom labels, lattice vectors and
        stores them before section_system and
        section_single_configuration_calculation are encountered.
        """
        # Checking whether topology, input
        # coordinates and trajectory files exist
        atLeastOneFileExist = False
        working_dir_name = os.path.dirname(os.path.abspath(self.fName))
        for k,v in self.fileDict.items():
            if k.startswith(PARSERTAG + '_inout_file'):
                if v.value:
                    file_name = os.path.normpath(os.path.join(working_dir_name, v.value))
                    self.fileDict[k].fileSupplied = os.path.isfile(file_name)
                    self.fileDict[k].activeInfo = False
                    if self.fileDict[k].fileSupplied:
                        self.fileDict[k].fileName = file_name
                        self.fileDict[k].activeInfo = True
                        if self.fileDict[k].activeInfo:
                            atLeastOneFileExist = True
                    else:
                        self.fileDict[k].value = None
        updateDict = {
            'startSection'   : [[PARSERTAG+'_section_input_output_files']],
            'dictionary'     : self.fileDict
            }
        self.metaStorage.update(updateDict)
        self.metaStorage.updateBackend(backend.superBackend,
                startsection=[PARSERTAG+'_section_input_output_files'],
                autoopenclose=False)
        if atLeastOneFileExist:
            self.MDData.initializeFileHandlers(self)
            if self.topology:
                self.topoDict = self.topology.topoDict
            if self.trajectory:
                self.atompositions = self.trajectory.positions()
            if self.inputcoords:
                self.inputpositions = self.inputcoords.positions()
            if self.outputcoords:
                self.outputpositions = self.outputcoords.positions()
        #self.MDiter += self.MDstep
        self.MDiter += 1
        self.MDcurrentstep = int(self.MDiter)
        self.stepcontrolDict.update({"MDcurrentstep" : int(self.MDcurrentstep)})
            #pass

    def onOpen_x_namd_section_control_parameters(self, backend, gIndex, section):
        # keep track of the latest section
        if self.inputControlIndex is None:
            self.inputControlIndex = gIndex

    def onClose_x_namd_section_control_parameters(self, backend, gIndex, section):
        section_control_Dict = {}
        section_control_Dict.update(self.cntrlDict)
        updateDict = {
            'startSection' : [[PARSERTAG+'_section_control_parameters']],
            'dictionary'   : section_control_Dict
            }
        self.metaStorage.update(updateDict)
        self.metaStorage.updateBackend(backend.superBackend,
                startsection=[PARSERTAG+'_section_control_parameters'],
                autoopenclose=False)
        nparms = 0
        nparmKey = isMetaStrInDict("PARAMETERS",self.cntrlDict)
        if nparmKey is not None:
            if self.cntrlDict[nparmKey].activeInfo:
                nparms = np.array(self.cntrlDict[nparmKey].value).shape[0]
        if nparms>0:
            backend.superBackend.addValue(PARSERTAG+"_inout_control_number_of_parameters", int(nparms))
        # NAMD prints the initial and final energies to the log file.
        # The total number of MD steps in NAMD is nsteps irrelevant
        # to the number of steps in log file of energy file (.edr)
        nsteps = 0
        nlogsteps = 0
        nstructstep = 0
        ninputstep = 0
        noutputstep = 0
        nbinsteps = 0
        ntrajsteps = 0
        nvelsteps = 0
        nforcesteps = 0
        nstructKey = isMetaStrInDict("STRUCTURE FILE",self.cntrlDict)
        ninputKey = isMetaStrInDict("COORDINATE PDB",self.cntrlDict)
        noutputKey = isMetaStrInDict("OUTPUT FILENAME",self.cntrlDict)
        nstepKey = isMetaStrInDict("NUMBER OF STEPS", self.cntrlDict)
        nlogKey = isMetaStrInDict("ENERGY OUTPUT STEPS", self.cntrlDict)
        nxoutKey = isMetaStrInDict("DCD FREQUENCY", self.cntrlDict)
        nvoutKey = isMetaStrInDict("VELOCITY DCD FREQUENCY", self.cntrlDict)
        nfoutKey = isMetaStrInDict("FORCE DCD FREQUENCY", self.cntrlDict)
        nbinoutKey = isMetaStrInDict("BINARY OUTPUT", self.cntrlDict)
        if nstepKey is not None:
            if self.cntrlDict[nstepKey].activeInfo:
                nsteps = conv_int(self.cntrlDict[nstepKey].value, default=0)
            else:
                nsteps = conv_int(self.cntrlDict[nstepKey].defaultValue, default=0)
        if nlogKey is not None:
            if self.cntrlDict[nlogKey].activeInfo:
                nlogsteps = conv_int(self.cntrlDict[nlogKey].value, default=1)
            else:
                nlogsteps = conv_int(self.cntrlDict[nlogKey].defaultValue, default=1)
        if nbinoutKey is not None:
            if self.cntrlDict[nbinoutKey].activeInfo:
                binsteps = [nsteps] if self.cntrlDict[
                        nbinoutKey].value == "FILES WILL BE USED" else []
            else:
                binsteps = []
        else:
            binsteps = []
        if nxoutKey is not None:
            if self.cntrlDict[nxoutKey].activeInfo:
                ntrajsteps = conv_int(self.cntrlDict[nxoutKey].value, default=0)
        if nstructKey is not None:
            if self.cntrlDict[nstructKey].activeInfo:
                nstructstep = 1
        if ninputKey is not None:
            if self.cntrlDict[ninputKey].activeInfo:
                ninputstep = 1
        if noutputKey is not None:
            if self.cntrlDict[noutputKey].activeInfo:
                noutputstep = 1
        if nvoutKey is not None:
            if self.cntrlDict[nvoutKey].activeInfo:
                nvelsteps = conv_int(self.cntrlDict[nvoutKey].value, default=0)
        if nfoutKey is not None:
            if self.cntrlDict[nfoutKey].activeInfo:
                nforcesteps = conv_int(self.cntrlDict[nfoutKey].value, default=0)

        if nlogsteps>0:
            logsteps = [i for i in range(0, nsteps, nlogsteps)]
            logsteps.append(nsteps)
        else:
            logsteps = [i for i in range(0, nsteps)]
            logsteps.append(nsteps)
        if ntrajsteps>0:
            trajsteps = [i for i in range(0, nsteps, ntrajsteps)]
            trajsteps.append(nsteps)
        else:
            trajsteps = []
            if(ninputstep>0 or nstructstep>0):
                trajsteps.append(0)
            if noutputstep>0:
                trajsteps.append(nsteps)
        if nvelsteps>0:
            velsteps = [i for i in range(0, nsteps, nvelsteps)]
            velsteps.append(nsteps)
        else:
            velsteps = []
            #velsteps.append(nsteps)
        if nforcesteps>0:
            forcesteps = [i for i in range(0, nsteps, nforcesteps)]
            forcesteps.append(nsteps)
        else:
            forcesteps = []
            #forcesteps.append(nsteps)
        self.stepcontrolDict.update({"logsteps"     : logsteps})
        self.stepcontrolDict.update({"nextlogsteps" : logsteps})
        self.stepcontrolDict.update({"trajsteps"    : trajsteps})
        self.stepcontrolDict.update({"velsteps"     : velsteps})
        self.stepcontrolDict.update({"forcesteps"   : forcesteps})
        self.stepcontrolDict.update({"binsteps"     : binsteps})
        steps = []
        for step in range(0,nsteps+1):
            if(step in logsteps or
               step in trajsteps or
               step in velsteps or
               step in forcesteps):
               steps.append(step)
        #steps = logsteps if len(logsteps)>len(trajsteps) else trajsteps
        #followsteps = "log" if len(logsteps)>len(trajsteps) else "traj"
        #steps = steps if len(steps)>len(velsteps) else velsteps
        #followsteps = followsteps if len(steps)>len(velsteps) else "vel"
        #self.stepcontrolDict.update({"steps"  : steps if len(steps)>len(forcesteps) else forcesteps})
        #self.stepcontrolDict.update({"follow" : followsteps if len(steps)>len(enersteps) else "force"})
        self.stepcontrolDict.update({"steps"  : steps})
        #self.stepcontrolDict.update({"follow" : followsteps})

    def onOpen_section_method(self, backend, gIndex, section):
        # keep track of the latest method section
        self.secMethodGIndex = gIndex
        if self.inputMethodIndex is None:
            self.inputMethodIndex = gIndex
        else:
            backend.openNonOverlappingSection("section_method_to_method_refs")
            backend.addValue("method_to_method_kind", "core_settings")
            backend.addValue("method_to_method_ref", self.inputMethodIndex)
            backend.closeNonOverlappingSection("section_method_to_method_refs")
        if self.mainMethodIndex is None:
            self.mainMethodIndex = gIndex

    def onClose_section_method(self, backend, gIndex, section):
        """Trigger called when section_method is closed.
        """
        # input method
        pass

    def onOpen_section_sampling_method(self, backend, gIndex, section):
        # keep track of the latest sampling description section
        self.secSamplingGIndex = gIndex
        #self.inputControlIndex = backend.superBackend.openSection(PARSERTAG + '_section_control_parameters')

    def onClose_section_sampling_method(self, backend, gIndex, section):
        """Trigger called when section_sampling_method is closed.

        Writes sampling method details for minimization and molecular dynamics.
        """
        # check control keywords were found throguh dictionary support
        #backend.superBackend.closeSection(PARSERTAG + '_section_control_parameters', self.inputControlIndex)
        section_sampling_Dict = get_updateDictionary(self, 'sampling')
        updateDict = {
            #'startSection' : [['section_sampling_method']],
            'startSection' : [['section_run']],
            'dictionary' : section_sampling_Dict
            }
        self.metaStorage.update(updateDict)
        self.metaStorage.updateBackend(backend.superBackend,
                startsection=['section_run'],
                #startsection=['section_sampling_method'],
                autoopenclose=False)

    def onOpen_section_topology(self, backend, gIndex, section):
        # keep track of the latest topology description section
        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            self.secTopologyGIndex = backend.superBackend.openSection("section_topology")
        else:
            self.secTopologyGIndex = gIndex

    def onClose_section_topology(self, backend, gIndex, section):
        """Trigger called when section_topology is closed.
        """
        section_topology_Dict = get_updateDictionary(self, 'topology')
        updateDict = {
            'startSection' : [
                ['section_topology']],
            #'muteSections' : [['section_sampling_method']],
            'dictionary' : section_topology_Dict
            }
        self.metaStorage.update(updateDict)
        self.metaStorage.updateBackend(backend.superBackend,
                startsection=['section_topology'],
                autoopenclose=False)
        self.topology_atom_type_and_interactions(backend, gIndex)

        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            backend.superBackend.closeSection("section_topology", self.secTopologyGIndex)

    def onOpen_section_system(self, backend, gIndex, section):
        # keep track of the latest system description section
        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            self.secSystemGIndex = backend.superBackend.openSection("section_system")
        else:
            self.secSystemGIndex = gIndex

    def onClose_section_system(self, backend, gIndex, section):
        """Trigger called when section_system is closed.

        Writes atomic positions, atom labels and lattice vectors.
        """
        # Our special recipe for the confused backend and more
        # See our other recipes at the back of this parser: MetaInfoStorage! :)
        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            SloppyBackend = backend.superBackend
        else:
            SloppyBackend = backend

        numatoms = None

        # NAMD default unit vectors. If all are zeros than the simulation cell is not periodic.
        unit_vectors = np.asarray(
                [[0.0,0.0,0.0],
                 [0.0,0.0,0.0],
                 [0.0,0.0,0.0]]
                )
        vaKey = isMetaStrInDict("PERIODIC CELL BASIS 1",self.cntrlDict)
        vbKey = isMetaStrInDict("PERIODIC CELL BASIS 2",self.cntrlDict)
        vcKey = isMetaStrInDict("PERIODIC CELL BASIS 3",self.cntrlDict)
        if vaKey is not None:
            if self.cntrlDict[vaKey].value is not None:
                unit_vectors[:,0]=np.asarray([ float(i) for i in self.cntrlDict[vaKey].value.split()])
        if vbKey is not None:
            if self.cntrlDict[vbKey].value is not None:
                unit_vectors[:,1]=np.asarray([ float(i) for i in self.cntrlDict[vbKey].value.split()])
        if vcKey is not None:
            if self.cntrlDict[vcKey].value is not None:
                unit_vectors[:,2]=np.asarray([ float(i) for i in self.cntrlDict[vcKey].value.split()])
        if self.trajectory is not None:
            if self.trajectory.unitcell_vectors is None:
                self.trajectory.unitcell_vectors = unit_vectors

        if self.topology:
            if (self.secTopologyGIndex is None or
                (self.secTopologyGIndex == -1 or
                self.secTopologyGIndex == "-1")):
                self.onOpen_section_topology(backend, None, None)
                self.onClose_section_topology(backend, None, None)
            SloppyBackend.addValue("topology_ref", self.secTopologyGIndex)

        coordinates=None
        steps = self.stepcontrolDict["steps"]
        if self.trajectory is not None:
            coordinates=self.trajectory
            positions=self.atompositions
        elif(self.inputcoords is not None and
             self.MDcurrentstep == steps[0]):
            coordinates=self.inputcoords
            positions=self.inputpositions
        elif(self.outputcoords is not None and
             self.MDcurrentstep == steps[-1]):
            coordinates=self.outputcoords
            positions=self.outputpositions

        if(coordinates is not None and positions is not None):
            self.trajRefSingleConfigurationCalculation = gIndex
            if coordinates.unitcell_vectors is not None:
                unit_cell = np.asarray(self.metaStorage.convertUnits(
                    coordinates.unitcell_vectors, "Angstrom", self.unitDict))
                SloppyBackend.addArrayValues('simulation_cell', unit_cell)
                SloppyBackend.addArrayValues('lattice_vectors', unit_cell)
            if coordinates.unitcell_lengths is not None:
                SloppyBackend.addArrayValues(PARSERTAG + '_lattice_lengths', coordinates.unitcell_lengths)
            if coordinates.unitcell_angles is not None:
                SloppyBackend.addArrayValues(PARSERTAG + '_lattice_angles', coordinates.unitcell_angles)
            # NAMD Guide says: (See http://www.ks.uiuc.edu/Research/namd/2.12/ug/node11.html)
            # Positions in PDB/NAMD binary/DCD files are stored in A.
            #pos_obj = getattr(self.trajectory, 'positions', None)
            #if callable(pos_obj):
            if(isinstance(positions, np.ndarray) or isinstance(positions, (tuple, list))):
                SloppyBackend.addArrayValues('atom_positions', np.transpose(np.asarray(
                    self.metaStorage.convertUnits(positions, "Angstrom", self.unitDict))))
            if coordinates.velocities is not None:
                # Velocities in PDB files are stored in A/ps units.(PDB files are read for input
                #     coordinates, velocities, and forces)
                # Velocities in NAMD binary/DCD files are stored in NAMD internal units and must be multiplied
                #     by PDBVELFACTOR=20.45482706 to convert to A/ps. (These files are used for output trajectory)
                SloppyBackend.addArrayValues('atom_velocities', np.transpose(np.asarray(
                    self.metaStorage.convertUnits(
                        coordinates.velocities, "20.45482706*Angstrom/(pico-second)", self.unitDict))))
                # need to transpose array since its shape is [number_of_atoms,3] in the metadata

            if self.topology is not None:
                atom_labels = self.topoDict["atom_element_list"]
                numatoms = len(atom_labels)
                SloppyBackend.addArrayValues('atom_labels', np.asarray(atom_labels))
                pass
        else:
            if unit_vectors is not None:
                unit_cell = np.asarray(self.metaStorage.convertUnits(
                    unit_vectors, "Angstrom", self.unitDict))
                SloppyBackend.addArrayValues('simulation_cell', unit_cell)
                SloppyBackend.addArrayValues('lattice_vectors', unit_cell)

        if self.trajectory is not None:
            # Read the next step at trajectory in advance
            # If iread returns None, it will be the last step
            try:
                self.atompositions = self.trajectory.positions()
                if self.atompositions is not None:
                    numatoms = len(self.atompositions)
                #self.MDiter += self.MDstep
                #self.MDiter += 1
            except AttributeError:
                self.atompositions = None
                pass

        if unit_vectors is not None:
            unit_periodicity=np.asarray([False,False,False])
            if np.linalg.norm(unit_vectors[0])>0.0:
                unit_periodicity[0]=True
            if np.linalg.norm(unit_vectors[1])>0.0:
                unit_periodicity[1]=True
            if np.linalg.norm(unit_vectors[2])>0.0:
                unit_periodicity[2]=True
            SloppyBackend.addArrayValues('configuration_periodic_dimensions', unit_periodicity)

        if numatoms:
            SloppyBackend.addValue("number_of_atoms", numatoms)

        if (gIndex is None or gIndex == -1 or gIndex == "-1"):
            SloppyBackend.closeSection("section_system", self.secSystemGIndex)

    def onOpen_section_single_configuration_calculation(self, backend, gIndex, section):
        # write the references to section_method and section_system
        self.singleConfCalcs.append(gIndex)
        self.secSingleGIndex = backend.superBackend.openSection("section_single_configuration_calculation")

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        """Trigger called when section_single_configuration_calculation is closed.

        Write number of steps in MD or Minimization.
        Check for convergence of geometry optimization.
        Write energy values at MD and with error in Minimization.
        Write reference to section_method and section_system
        """

        self.lastCalculationGIndex = gIndex
        steps = self.stepcontrolDict["steps"]
        logsteps = self.stepcontrolDict["logsteps"]
        velsteps = self.stepcontrolDict["velsteps"]
        forcesteps = self.stepcontrolDict["forcesteps"]
        trajsteps = self.stepcontrolDict["trajsteps"]
        #print("PRINTING steps:",steps)
        #print("PRINTING logsteps:",logsteps)
        #print("PRINTING velsteps:",velsteps)
        #print("PRINTING forcesteps:",forcesteps)
        #print("PRINTING trajsteps:",trajsteps)
        #if len(steps)>1:
        #    self.MDstep = steps[1]-steps[0]
        #if self.MDiter<len(steps):
        #    self.MDcurrentstep = steps[self.MDiter]
        #else:
        #    self.MDcurrentstep += 1
        #if self.MDiter+1<len(steps):
        #    self.MDnextstep = steps[self.MDiter+1]
        #    self.stepcontrolDict.update({"nextlogsteps":logsteps[self.MDiter+1:]})
        #else:
        #    self.MDnextstep = steps[-1] + 1
        #self.stepcontrolDict.update({"MDcurrentstep" : self.MDcurrentstep})

        if self.MDcurrentstep in logsteps:
            section_frameseq_Dict = get_updateDictionary(self, 'frameseq')
            updateFrameDict = {
                'startSection' : [
                    ['section_frame_sequence']],
                #'muteSections' : [['section_sampling_method']],
                'dictionary' : section_frameseq_Dict
                }
            self.metaStorage.update(updateFrameDict)
            section_singlevdw_Dict = get_updateDictionary(self, 'singlevdw')
            updateDictVDW = {
                'startSection' : [
                    ['section_energy_van_der_Waals']],
                #'muteSections' : [['section_sampling_method']],
                'dictionary' : section_singlevdw_Dict
                }
            self.secVDWGIndex = backend.superBackend.openSection("section_energy_van_der_Waals")
            self.metaStorage.update(updateDictVDW)
            self.metaStorage.updateBackend(backend.superBackend,
                    startsection=['section_energy_van_der_Waals'],
                    autoopenclose=False)
            backend.superBackend.closeSection("section_energy_van_der_Waals", self.secVDWGIndex)
            section_singlecalc_Dict = get_updateDictionary(self, 'singleconfcalc')
            updateDict = {
                'startSection' : [
                    ['section_single_configuration_calculation']],
                #'muteSections' : [['section_sampling_method']],
                'dictionary' : section_singlecalc_Dict
                }
            self.metaStorage.update(updateDict)
            self.metaStorage.updateBackend(backend.superBackend,
                    startsection=['section_single_configuration_calculation'],
                    autoopenclose=False)
        if self.MDcurrentstep in forcesteps:
            pos_obj = getattr(self.trajectory, 'positions', None)
            #if callable(pos_obj):
            if self.trajectory.forces is not None:
                # Forces in PDB/NAMD binary/DCD files are stored in kcal/mol/A
                SloppyBackend.addArrayValues('atom_forces', np.transpose(np.asarray(
                    self.metaStorage.convertUnits(
                        self.atompositions.velocities, "kcal/(mol*Angstrom)", self.unitDict))))
                # need to transpose array since its shape is [number_of_atoms,3] in the metadata

        if(self.MDcurrentstep in trajsteps or
           self.MDcurrentstep in velsteps):
            #print("PRINTING traj steps:",self.MDcurrentstep,trajsteps)
            self.onOpen_section_system(backend, None, None)
            self.onClose_section_system(backend, None, None)
            backend.addValue('single_configuration_calculation_to_system_ref', self.secSystemGIndex)
            self.MDiter += 1
        else:
            if(self.MDcurrentstep in logsteps or
               self.MDcurrentstep in forcesteps):
                self.MDiter += 1
            #if((self.MDcurrentstep in logsteps and
            #    self.MDiter+1 in steps) or
            #    (self.MDcurrentstep in forcesteps and
            #    self.MDiter+1 in steps)):
            #    self.MDiter += 1
        #if self.MDiter<len(steps):
        #    self.MDcurrentstep = steps[self.MDiter]
        #else:
        #    self.MDcurrentstep += 1
        if self.MDiter+1<len(steps):
            self.MDnextstep = int(steps[self.MDiter])
            self.stepcontrolDict.update({"nextlogsteps":logsteps[self.MDiter:]})
        else:
            self.MDnextstep = int(steps[-1] + 1)
        self.stepcontrolDict.update({"MDcurrentstep" : int(self.MDcurrentstep)})
        #print("PRINTING section single MDiter:",self.MDiter)
        backend.superBackend.closeSection("section_single_configuration_calculation", self.secSingleGIndex)

    def setStartingPointCalculation(self, parser):
        backend = parser.backend
        backend.openSection('section_calculation_to_calculation_refs')
        if self.lastCalculationGIndex:
            backend.addValue('calculation_to_calculation_ref', self.lastCalculationGIndex)
            backend.closeSection('section_calculation_to_calculation_refs')
        return None

    def fileNameTrans(self, fname, value):
        fileNameTrans = {
                "output_coord" : ".coor",
                "output_vel"   : ".vel",
                "output_xsc"   : ".xsc",
                }
        for k,v in fileNameTrans.items():
            if fname in k:
                return value+v
        return value

    def mdInfoTrans(self, fname, value):
        mdInfoDict = {
                "minimization" : [],
                "thermostat"   : [],
                "barostat"     : []
                }
        cgmin = isMetaStrInDict("MINIMIZATION",self.cntrlDict)
        quench = isMetaStrInDict("VELOCITY QUENCHING",self.cntrlDict)
        tcouple = isMetaStrInDict("TEMPERATURE COUPLING",self.cntrlDict)
        trescale = isMetaStrInDict("VELOCITY RESCALE FREQ",self.cntrlDict)
        treassign = isMetaStrInDict("VELOCITY REASSIGNMENT FREQ",self.cntrlDict)
        tandersen = isMetaStrInDict("LOWE-ANDERSEN DYNAMICS",self.cntrlDict)
        tlangevin = isMetaStrInDict("LANGEVIN DYNAMICS",self.cntrlDict)
        pberendsen = isMetaStrInDict("BERENDSEN PRESSURE COUPLING",self.cntrlDict)
        plangevin = isMetaStrInDict("LANGEVIN PISTON PRESSURE CONTROL",self.cntrlDict)
        mini = None
        thermo = None
        baro = None
        if cgmin is not None:
            if self.cntrlDict[cgmin].value is not None:
                mini = 'cg'
        if quench is not None:
            if self.cntrlDict[quench].value is not None:
                mini = 'quench'
        if tcouple is not None:
            if self.cntrlDict[tcouple].value is not None:
                thermo = 'tcouple'
        if trescale is not None:
            if self.cntrlDict[trescale].value is not None:
                thermo = 'rescale'
        if treassign is not None:
            if self.cntrlDict[treassign].value is not None:
                thermo = 'reassign'
        if tandersen is not None:
            if self.cntrlDict[tandersen].value is not None:
                thermo = 'andersen'
        if tlangevin is not None:
            if self.cntrlDict[tlangevin].value is not None:
                thermo = 'langevin'
        if pberendsen is not None:
            if self.cntrlDict[pberendsen].value is not None:
                baro = 'berendsen'
        if plangevin is not None:
            if self.cntrlDict[plangevin].value is not None:
                baro = 'langevin'
        for k,v in mdInfoDict.items():
            if fname in k:
                if 'thermostat' in fname:
                    return thermo
                elif 'barostat' in fname:
                    return baro
                elif 'minimization' in fname:
                    return mini
                else:
                    return value
        return value

    def convertInt(self, fname, value):
        keyMapper = {
                "TS" : "MDcurrentstep",
                "DCDSTEP" : "MDcurrentstep",
                "VELSTEP" : "MDcurrentstep",
                "FORCESTEP" : "MDcurrentstep"
                }
        if value is not None:
            for k,v in keyMapper.items():
                if fname in k:
                    return int(value)
        return value

    def updateTime(self, fname, value):
        keyMapper = {
                "TIME" : "MDcurrentstep",
                }
        timestep = 1.0
        istimestep = isMetaStrInDict("TIMESTEP",self.cntrlDict)
        if istimestep is not None:
            if self.cntrlDict[istimestep].value is not None:
                timestep = float(self.cntrlDict[istimestep].value)
        if value is not None:
            for k,v in keyMapper.items():
                if fname in k:
                    return float(value)*timestep
        return value

    def build_subMatchers(self):
        """Builds the sub matchers to parse the main output file.
        """
        mddataNameList=getList_MetaStrInDict(self.mddataDict)

        mdoutSubParsers = [
            # timestep header Parser
            { "startReStr"        : r"\s*ETITLE:\s*",
              "parser"            : "table_parser",
              "parsername"        : "header_parser",
              "waitlist"          : None,
              "stopOnMatchStr"    : r"(?:^\s*$|\s*ENERGY:\s*)",
              "quitOnMatchStr"    : r"^====================================================",
              "metaNameStart"     : PARSERTAG + "_mdout_",
              "matchNameList"     : mddataNameList,
              "matchNameDict"     : "mddataDict",
              "updateMatchDict"   : True,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : False,
              "parserOptions"     : {
                  "header"           : True,
                  "headersave"       : "mdoutHeaderDict",
                  "wrap"             : False,
                  "tablelines"       : 0,
                  "tablestartsat"    : r"\s*ETITLE:\s*",
                  "tableendsat"      : r"^\s*$",
                  "lineFilter"       : None,
                  "movetostopline"   : False,
                  "parsercntrlattr"  : "MDcurrentstep",
                  "parsercntrlin"    : "logsteps",
                  "lookupdict"       : "stepcontrolDict"
                  }
              },
            # timestep energy outputs Parser
            { "startReStr"        : r"\s*ENERGY:\s*",
              "parser"            : "table_parser",
              "parsername"        : "log_step_parser",
              "waitlist"          : None,
              "stopOnMatchStr"    : r"(?:^\s*$|\s*ENERGY:\s*)",
              "quitOnMatchStr"    : r"^====================================================",
              "metaNameStart"     : PARSERTAG + "_mdout_",
              "matchNameList"     : mddataNameList,
              "matchNameDict"     : "mddataDict",
              "updateMatchDict"   : True,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : False,
              "parserOptions"     : {
                  "header"           : False,
                  "headerList"       : "mdoutHeaderDict",
                  "wrap"             : False,
                  "tablelines"       : 0,
                  "tablestartsat"    : r"\s*ENERGY:\s*",
                  "tableendsat"      : r"^\s*$",
                  "lineFilter"       : None,
                  "movetostopline"   : False,
                  "parsercntrlattr"  : "MDcurrentstep",
                  "parsercntrlin"    : "logsteps",
                  "lookupdict"       : "stepcontrolDict"
                  }
              },
            # trajectory Parser
            { "startReStr"        : r"\s*WRITING\s*COORDINATES\s*TO\s*DCD\s*FILE\s*",
              "parser"            : "namelist_parser",
              "parsername"        : "dcd_parser",
              "stopOnMatchStr"    : r"\s*(?:ENERGY|TIMING|FINISHED|OPENING|The last)\s*",
              "quitOnMatchStr"    : r"\s*(?:ENERGY|TIMING|FINISHED|OPENING|The last)\s*",
              "metaNameStart"     : PARSERTAG + "_mdout_",
              "matchNameList"     : mddataNameList,
              "matchNameDict"     : "mddataDict",
              "updateMatchDict"   : False,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : True,
              "parserOptions"     : {
                  "controlsections"  : ["section_single_configuration_calculation"],
                  "controlsave"      : "sectioncontrol",
                  "controldict"      : "stepcontrolDict",
                  }
              },
            # velocities Parser
            { "startReStr"        : r"\s*WRITING\s*VELOCITIES\s*TO\s*DCD\s*FILE\s*",
              "parser"            : "namelist_parser",
              "parsername"        : "veldcd_parser",
              "stopOnMatchStr"    : r"\s*(?:ENERGY|TIMING|FINISHED|OPENING|The last)\s*",
              "quitOnMatchStr"    : r"\s*(?:ENERGY|TIMING|FINISHED|OPENING|The last)\s*",
              "metaNameStart"     : PARSERTAG + "_mdout_",
              "matchNameList"     : mddataNameList,
              "matchNameDict"     : "mddataDict",
              "updateMatchDict"   : False,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : True,
              "parserOptions"     : {
                  "controlsections"  : ["section_single_configuration_calculation"],
                  "controlsave"      : "sectioncontrol",
                  "controldict"      : "stepcontrolDict",
                  }
              },
            # forces Parser
            { "startReStr"        : r"\s*WRITING\s*FORCES\s*TO\s*DCD\s*FILE\s*",
              "parser"            : "namelist_parser",
              "parsername"        : "forcedcd_parser",
              "stopOnMatchStr"    : r"\s*(?:ENERGY|TIMING|FINISHED|OPENING|The last)\s*",
              "quitOnMatchStr"    : r"\s*(?:ENERGY|TIMING|FINISHED|OPENING|The last)\s*",
              "metaNameStart"     : PARSERTAG + "_mdout_",
              "matchNameList"     : mddataNameList,
              "matchNameDict"     : "mddataDict",
              "updateMatchDict"   : False,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : True,
              "parserOptions"     : {
                  "controlsections"  : ["section_single_configuration_calculation"],
                  "controlsave"      : "sectioncontrol",
                  "controldict"      : "stepcontrolDict",
                  }
              },
            # thermostat save Parser
            { "startReStr"        : "AlwaysRun",
              "parser"            : "dictionary_parser",
              "parsername"        : "frameseq_step_parser",
              "waitlist"          : None,
              "stopOnMatchStr"    : "AlwaysStop",
              "quitOnMatchStr"    : "AlwaysStop",
              "metaNameStart"     : PARSERTAG + "_mdout_",
              "matchNameList"     : mddataNameList,
              "matchNameDict"     : "mddataDict",
              "updateMatchDict"   : True,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : True,
              "parserOptions"     : {
                  "dictionary"       : "stepcontrolDict",
                  "dicttype"         : "standard", # (standard or smartparser)
                  "readwritedict"    : "write",
                  "keyMapper"        : {"TS" : "MDcurrentstep",
                                        "DCDSTEP" : "MDcurrentstep",
                                        "VELSTEP" : "MDcurrentstep",
                                        "FORCESTEP" : "MDcurrentstep"},
                  "updatefunc"       : "max",
                  "updateattrs"      : ["MDcurrentstep"],
                  "controlsections"  : ["section_single_configuration_calculation"],
                  "controlsave"      : "sectioncontrol",
                  "controldict"      : "stepcontrolDict",
                  "controlattrs"     : ["MDcurrentstep"],
                  "preprocess"       : self.convertInt,
                  "postprocess"      : self.convertInt,
                  #"parsercntrlattr"  : "MDcurrentstep",
                  #"parsercntrlin"    : "steps",
                  #"lookupdict"       : "stepcontrolDict"
                  }
              },
            # time save Parser
            { "startReStr"        : "AlwaysRun",
              "parser"            : "dictionary_parser",
              "parsername"        : "frameseq_step_parser",
              "waitlist"          : None,
              "stopOnMatchStr"    : "AlwaysStop",
              "quitOnMatchStr"    : "AlwaysStop",
              "metaNameStart"     : PARSERTAG + "_mdout_",
              "matchNameList"     : mddataNameList,
              "matchNameDict"     : "mddataDict",
              "updateMatchDict"   : True,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : True,
              "parserOptions"     : {
                  "dictionary"       : "mddataDict",
                  "dicttype"         : "smartparser", # (standard or smartparser)
                  "readwritedict"    : "read",
                  "keyMapper"        : {"TIME" : "TS"},
                  "preprocess"       : self.convertInt,
                  "postprocess"      : self.updateTime,
                  }
              },
            # Section Control Parser
            { "startReStr"        : "AlwaysRun",
              "parser"            : "section_control_parser",
              "parsername"        : "section_parser",
              "waitlist"          : [["log_step_parser"],
                                     ["forcedcd_parser"],
                                     ["veldcd_parser"],
                                     ["dcd_parser"]],
              "stopOnMatchStr"    : r"\s*ENERGY:\s*",
              "quitOnMatchStr"    : r"\s*ENERGY:\s*",
              "metaNameStart"     : None,
              "matchNameList"     : None,
              "matchNameDict"     : None,
              "updateMatchDict"   : False,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : True,
              "parserOptions"     : {
                  "sectionname"      : "section_single_configuration_calculation",
                  "sectionopen"      : True,
                  "sectionopenattr"  : "MDcurrentstep",
                  "sectionopenin"    : "steps",
                  "sectionclose"     : True,
                  "sectioncloseattr" : "MDcurrentstep",
                  "sectionclosein"   : "steps",
                  "activatesection"  : "sectioncontrol",
                  "lookupdict"       : "stepcontrolDict"
                  }
              },
            # Step Control Parser
            { "startReStr"        : "AlwaysRun",
              "parser"            : "readline_control_parser",
              "parsername"        : "readline_parser",
              "waitlist"          : None,
              "stopOnMatchStr"    : "AlwaysStop",
              "quitOnMatchStr"    : "AlwaysStop",
              "metaNameStart"     : None,
              "matchNameList"     : None,
              "matchNameDict"     : None,
              "updateMatchDict"   : False,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : True,
              "parserOptions"     : {
                  "waitatlineStr"    : r"\s*ENERGY:\s*",
                  "controlwait"      : None,
                  #"controlwait"      : "nextlogsteps",
                  "controlattr"      : "MDcurrentstep",
                  "controlnextattr"  : "MDnextstep",
                  "controllast"      : -1,
                  "controlskip"      : [],
                  "controlin"        : "steps",
                  "controlcounter"   : "targetstep",
                  "controldict"      : "stepcontrolDict",
                  "lookupdict"       : "stepcontrolDict"
                  }
              }
            ]

        cntrlNameList=getList_MetaStrInDict(self.metaDicts['cntrl'])
        fileNameList=getList_MetaStrInDict(self.metaDicts['file'])
        extraNameList=getList_MetaStrInDict(self.metaDicts['extra'])

        mdinoutSubParsers = [
            # cntrl Parser
            { "startReStr"        : r"\s*Info:\s*SIMULATION\s*PARAMETERS:\s*",
              "parser"            : "namelist_parser",
              "parsername"        : "cntrl_parser",
              "stopOnMatchStr"    : r"\s*Info:\s*(?:SUMMARY\s*OF\s*PARAMETERS:|"
                                     "STRUCTURE\s*SUMMARY:)",
              "quitOnMatchStr"    : None,
              "metaNameStart"     : PARSERTAG + "_inout_control_",
              "matchNameList"     : cntrlNameList,
              "matchNameDict"     : "cntrlDict",
              "updateMatchDict"   : True,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : False,
              "parserOptions"     : { "lineFilter" : None }
              },
            # simplify md info Parser
            { "startReStr"        : r"\s*Info:\s*(?:SUMMARY\s*OF\s*PARAMETERS:|"
                                     "STRUCTURE\s*SUMMARY:)",
              "parser"            : "dictionary_parser",
              "parsername"        : "info_add_parser",
              "quitOnMatchStr"    : None,
              "stopOnMatchStr"    : "AlwaysStop",
              "quitOnMatchStr"    : "AlwaysStop",
              "metaNameStart"     : PARSERTAG + "_mdin_method_",
              "matchNameList"     : extraNameList,
              "matchNameDict"     : "extraDict",
              "updateMatchDict"   : True,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : True,
              "parserOptions"     : {
                  "dictionary"       : "cntrlDict",
                  "dicttype"         : "smartparser", # (standard or smartparser)
                  "readwritedict"    : "read",
                  "keyMapper"        : {
                      "minimization" : "MINIMIZATION",
                      "minimization" : "VELOCITY QUENCHING",
                      "integrator"   : "VERLET INTEGRATOR",
                      "thermostat"   : "TEMPERATURE COUPLING",
                      "thermostat"   : "VELOCITY RESCALE FREQ",
                      "thermostat"   : "VELOCITY REASSIGNMENT FREQ",
                      "thermostat"   : "LOWE-ANDERSEN DYNAMICS",
                      "thermostat"   : "LANGEVIN DYNAMICS",
                      "barostat"     : "BERENDSEN PRESSURE COUPLING",
                      "barostat"     : "LANGEVIN PISTON PRESSURE CONTROL"
                      },
                  "preprocess"       : self.mdInfoTrans,
                  "postprocess"      : self.mdInfoTrans
                  }
              }
            ]

        fileinSubParsers = [
            # file Parser
            { "startReStr"        : "AlwaysRun",
              "parser"            : "dictionary_parser",
              "parsername"        : "file_parser",
              "waitlist"          : None,
              "stopOnMatchStr"    : "AlwaysStop",
              "quitOnMatchStr"    : "AlwaysStop",
              "metaNameStart"     : PARSERTAG + "_inout_file",
              "matchNameList"     : fileNameList,
              "matchNameDict"     : "fileDict",
              "updateMatchDict"   : True,
              "onlyCaseSensitive" : True,
              "stopOnFirstLine"   : True,
              "parserOptions"     : {
                  "dictionary"       : "cntrlDict",
                  "dicttype"         : "smartparser", # (standard or smartparser)
                  "readwritedict"    : "read",
                  "keyMapper"        : {
                      "structure"    : "STRUCTURE FILE",
                      "traj_coord"   : "Info: DCD FILENAME",
                      "traj_vel"     : "VELOCITY DCD FILENAME",
                      "traj_force"   : "FORCE DCD FILENAME",
                      "input_coord"  : "COORDINATE PDB",
                      "output_coord" : "OUTPUT FILENAME",
                      },
                  "preprocess"       : self.fileNameTrans
                  }
              }
            ]

        ########################################
        # return main Parser
        return [
            SM(name='NewRun',
                startReStr=r"\s*Info:\s*NAMD\s*[0-9.]+\s*for\s*",
                endReStr=r"\s*End\s*of\s*program\s*",
                repeats=True,
                required=True,
                forwardMatch=True,
                sections=['section_run'],
                fixedStartValues={'program_name': PROGRAMNAME},
                subMatchers=[
                    SM(name='ProgramInfo',
                       startReStr=r"\s*Info:\s*NAMD\s*"
                                   "(?P<program_version>[0-9.]+)\s*for\s*"
                                   "(?P<"+PARSERTAG+"_build_osarch>.+)\s*"),
                    SM(name='license',
                       startReStr=r"\s*Info:\s*Please\s*visit\s*",
                       coverageIgnore=True,
                       adHoc=lambda p:
                       self.adHoc_read_store_text_stop_parsing(p,
                           stopOnMatchStr=r"\s*Info:\s*(?:Info:Based|"
                                           "Built|Running|Configuration)\s*",
                           quitOnMatchStr=None,
                           metaNameStart=PARSERTAG+"_",
                           metaNameStore=PARSERTAG+"_program_citation",
                           matchNameList=None,
                           matchNameDict=None,
                           onlyCaseSensitive=True,
                           stopOnFirstLine=False,
                           storeFirstLine=True,
                           storeStopQuitLine=True,
                           onQuitRunFunction=lambda p: p.backend.addValue(
                               PARSERTAG+"_program_citation",
                               p.lastMatch[
                                   PARSERTAG+"_program_citation"
                                   ].replace('\n', ' ').replace('Info:', ' ')
                               )
                           )
                       ),
                    SM(name='loghostinfo',
                       startReStr=r"\s*Info:\s*Built\s*"
                                   "(?P<"+PARSERTAG+"_program_build_date>"
                                   "[a-zA-Z0-9:. ]+)\s*by"),
                    SM(name='SectionControlParm',
                       startReStr=r"\s*Info:\s*SIMULATION\s*PARAMETERS\s*",
                       endReStr=r"\s*Info:\s*(?:SUMMARY\s*OF\s*PARAMETERS|"
                                 "STRUCTURE\s*SUMMARY|Entering\s*startup)\s*",
                       forwardMatch=True,
                       sections=['section_sampling_method', PARSERTAG + '_section_control_parameters'],
                       adHoc=lambda p:
                       self.adHoc_takingover_parsing(p,
                           stopOnMatchStr=r"\s*Info:\s*(?:SUMMARY\s*OF\s*PARAMETERS|"
                                           "STRUCTURE\s*SUMMARY|Entering\s*startup)",
                           quitOnMatchStr=None,
                           stopControl="stopControl", # if None or True, stop with quitMatch, else wait
                           record=False, # if False or None, no record, no replay
                           replay=0, # if 0 or None= no replay, if <0 infinite replay
                           parseOnlyRecorded=False, # if True, parsers only work on record
                           ordered=False,
                           subParsers=mdinoutSubParsers)
                       ), # END SectionControlParm
                    SM(name='FileNameCopy',
                       startReStr=r"\s*Info:\s*Finished\s*startup",
                       forwardMatch=True,
                       sections=[PARSERTAG+'_section_input_output_files'],
                       coverageIgnore=True,
                       adHoc=lambda p:
                       self.adHoc_takingover_parsing(p,
                           stopOnMatchStr=r"^\s*$",
                           quitOnMatchStr=r"^\s*$",
                           stopControl="stopControl", # if None or True, stop with quitMatch, else wait
                           record=False, # if False or None, no record, no replay
                           replay=0, # if 0 or None= no replay, if <0 infinite replay
                           parseOnlyRecorded=False, # if True, parsers only work on record
                           ordered=False,
                           subParsers=fileinSubParsers),
                       ),
                    SM(name='SingleConfigurationCalculationWithSystemDescription',
                       startReStr=r"\s*ETITLE:\s*TS\s*",
                       endReStr=r"(?:^===================================================="
                                 "|WallClock:|End\s*of\s*program)\s*",
                       forwardMatch=True,
                       subMatchers=[
                           # the actual section for a single configuration calculation starts here
                           SM(name='SingleConfigurationCalculation',
                              startReStr=r"\s*ETITLE:\s*TS\s*",
                              endReStr=r"(?:^===================================================="
                                        "|^WallClock:|End\s*of\s*program)\s*",
                              repeats=True,
                              forwardMatch=True,
                              #sections=['section_single_configuration_calculation'],
                              subMatchers=[
                                  SM(name='MDStep',
                                     startReStr=r"\s*ETITLE:\s*TS\s*"
                                                 "(?P<"+PARSERTAG+"_mdin_finline>.*)(?:'|\")?\s*,?",
                                     forwardMatch=True,
                                     adHoc=lambda p:
                                     self.adHoc_takingover_parsing(p,
                                         stopOnMatchStr=r"^====================================================",
                                         quitOnMatchStr=r"^====================================================",
                                         stopControl="stopControl", # if None or True, stop with quitMatch, else wait
                                         record=False, # if False or None, no record, no replay
                                         replay=0, # if 0 or None= no replay, if <0 infinite replay
                                         parseOnlyRecorded=False, # if True, parsers only work on record
                                         ordered=False,
                                         onlySubParsersReadLine=True,
                                         subParsers=mdoutSubParsers)
                                     )]
                              ) # END SingleConfigurationCalculation
                       ]), # END SingleConfigurationCalculationWithSystemDescription
                    # summary of computation
                    SM(name='ComputationTimings',
                       startReStr=r"^WallClock:",
                       subMatchers=[
                           SM(r"^WallClock:\s*[0-9:.eEdD]+\s*CPUTime:\s*[0-9.:eEdD]+\s*Memory:\s*[0-9:.KMGTB ]+"),
                       ]), # END Computation
                    SM(name='end_run',
                       startReStr=r"\s*End\s*of\s*program\s*"),
                ]) # END NewRun
            ]


class NamdParserInterface():
   """ A proper class envolop for running this parser from within python. """
   def __init__(self, backend, **kwargs):
       self.backend_factory = backend

   def parse(self, mainfile):
        from unittest.mock import patch
        logging.info('namd parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("namd.nomadmetainfo.json")
        parserInfo = {'name': 'namd-parser', 'version': '1.0'}
        context = NAMDParser()
        context.coverageIgnore = re.compile(r"^(?:" + r"|".join(context.coverageIgnoreList) + r")$")
        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                mainFileDescription=context.mainFileDescription(),
                metaInfoEnv=None,
                parserInfo=parserInfo,
                cachingLevelForMetaName=context.cachingLevelForMetaName,
                superContext=context,
                superBackend=backend)
        return backend


if __name__ == "__main__":
    parser = NAMDParser()
    parser.parse()
